import numpy as np
import loglikelihood as logl

#remove:
import pickle

STPMX=100.0 
import time
class LineSearch:
	def __init__(self, x, comput, panel, step = 1):
		self.alf = 1.0e-3     #Ensures sufficient decrease in function value.
		self.tolx = 1.0e-14  #Convergence criterion on fx.		
		self.step = step
		self.stpmax = STPMX * max((abs(np.sum(x**2)	))**0.5,len(x))
		self.comput = comput
		self.panel = panel

	def lnsrch(self, x, f, g, dx):
		
		#(x, f, g, dx) 

		self.check=0
		self.g = g
		self.msg = ""
		n=len(x)
		self.rev = False
		if f is None:
			raise RuntimeError('f cannot be None')
	
		summ=np.sum(dx**2)**0.5
		if summ > self.stpmax:
			dx = dx*self.stpmax/summ 
		slope=np.sum(g*dx)					#Scale if attempted step is too big.
		if slope <= 0.0:
			self.msg = "Roundoff problem"
			dx=-dx
			slope=np.sum(g*dx)
			self.rev = True
		test=0.0 															#Compute lambda min.
		for i in range(0,n): 
			temp=abs(dx[i])/max(abs(x[i]),1.0) 
			if (temp > test): test=temp 
		alamin = self.tolx/test 
		#*******CUSTOMIZATION
		#multithread:

		for i in range(1000):#Setting alam so that the largest step is valid. Set func to return None when input is invalid
			self.alam = 0.5**i*self.step #Always try full Newton step first.
			self.x = x + self.alam * dx
			self.f, self.ll = self.func(self.x) 
			if self.f != None: break
		#*************************
		f2=0
		alam2 = self.alam
		alamstart = self.alam#***********CUSTOMIZATION
		max_iter = 1000
		for self.k in range (0,max_iter):			#Start of iteration loop.
			self.x = x + self.alam * dx			
			if self.k > 0: 
				self.f, self.ll = self.func(self.x) 
			if self.f is None:
				print('The function returned None')
				self.f = f
			if (self.alam < alamin):   #Convergence on delta x. For zero finding,the calling program should verify the convergence.
				self.x = x*1 
				self.check = 1
				self.f = f
				self.msg = "Convergence on delta dx"
				return
			elif (self.f >= f+self.alf*self.alam*slope): 
				self.msg = "Sufficient function increase"
				#print(f'LL single:{self.ll.LL}')
				#self.ll = self.ll2
				return							#Sufficient function increase
			else:  															#Backtrack.
				if (self.alam == alamstart):#***********CUSTOMIZATION  alam == 1.0
					tmplam = -slope/(2.0*(self.f-f-slope))  	#First time.
				else:  														#Subsequent backtracks.
					rhs1 = self.f-f-self.alam*slope 
					rhs2 = f2-f-alam2*slope 
					a=(rhs1/(self.alam**2)-rhs2/(alam2*alam2))/(self.alam-alam2) 
					b=(-alam2*rhs1/(self.alam**2)+self.alam*rhs2/(alam2*alam2))/(self.alam-alam2) 
					if (a == 0.0):
						tmplam = -slope/(2.0*b)  
					else:  
						disc=b*b-3.0*a*slope 
						if (disc < 0.0):
							tmplam = 0.5*self.alam  
						elif (b >= 0.0):
							tmplam=-(b+(disc)**0.5)/(3.0*a) 
						else:
							tmplam=slope/(-b+(disc)**0.5)
					if (tmplam > 0.5*self.alam): 
						tmplam = 0.5*self.alam   								#  lambda<=0.5*lambda1
			alam2 = self.alam 
			f2 = self.f
			self.alam = max(tmplam, 0.1*self.alam)								#lambda>=0.1*lambda1
			if alamstart<1.0:#*************CUSTOMIZATION
				self.alam = min((self.alam, alamstart*0.9**self.k))
				
			self.msg = f"No function increase after {max_iter} iterations"

	def func(self,x):
		if False:#for debug
			f = open('ctest.dump','wb')
			pickle.dump((x, self.panel, self.comput.constr), f)
			f.close()		
		ll = logl.LL(x, self.panel, self.comput.constr)
		if ll is None:
			return None, None
		elif ll.LL is None:
			return None, None
		return ll.LL, ll

	


#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
import stat_functions as stat


def get(g, x, H, constr, f, hessin, ev_constr, simple=True):
	n = len(x)
	if simple or (H is None):
		dx = -(np.dot(hessin,g.reshape(n,1))).flatten()
	else:
		#dx, H =solve(self.H, self.g, args.args_v, self.constr, f)
		dx, H = solve_delete(constr, H, g, x, f, ev_constr)	
	dx_norm = normalize(dx, x)
	
	return dx, dx_norm, H
	


def solve(H, g, x, constr, f, ev_constr):
	"""Solves a second degree taylor expansion for the dc for df/dc=0 if f is quadratic, given gradient
	g, hessian H, inequalty constraints c and equalitiy constraints c_eq and returns the solution and 
	and index constrained indicating the constrained variables"""
	if H is None:
		raise RuntimeError('Hessian is None')
	

	dx_init, H = linalg_solve(H, g, f, x, ev_constr)
	if constr is None:
		return dx_init	
	n=len(H)
	k=len(constr)
	H=np.concatenate((H,np.zeros((n,k))),1)
	H=np.concatenate((H,np.zeros((k,n+k))),0)
	g=np.append(g,(k)*[0])
	
	for i in range(k):
		H[n+i,n+i]=1
	j=0
	dx=np.zeros(len(g))
	for i in constr.fixed:
		#kuhn_tucker(constr.fixed[i],i,j,n, H, g, x,dx, recalc=False)
		kuhn_tucker2(constr.fixed[i],i,j,n, H, g, x, f,dx, dx_init, ev_constr, recalc=False)
		j+=1
	dx, H = linalg_solve(H, g, f, x, ev_constr)	
	OK=False
	w=0
	for r in range(50):
		j2=j
		
		for i in constr.intervals:
			dx, H=kuhn_tucker(constr.intervals[i],i,j2,n, H, g, x, f,dx, ev_constr)
			j2+=1
		OK=constr.within(x+dx[:n],False)
		if OK: 
			break
		if r==k+3:
			#print('Unable to set constraints in computation calculation')
			break

	return dx[:n], H


def solve_delete(constr,H, g, x, f, ev_constr):
	"""Solves a second degree taylor expansion for the dc for df/dc=0 if f is quadratic, given gradient
	g, hessian H, inequalty constraints c and equalitiy constraints c_eq and returns the solution and 
	and index constrained indicating the constrained variables"""
	
	if H is None:
		raise RuntimeError('Cant solve with no coefficient matrix')
	try:
		list(constr.keys())[0]
	except:
		dx, H = linalg_solve(H, g, f, x, ev_constr)
		return dx
	
	H_orig = np.array(H)
	m=len(H)
	
	idx=np.ones(m,dtype=bool)
	delmap=np.arange(m)
	if len(list(constr.fixed.keys()))>0:#removing fixed constraints from the matrix
		idx[list(constr.fixed.keys())]=False
		H=H[idx][:,idx]
		g=g[idx]
		delmap-=np.cumsum(idx==False)
		delmap[idx==False]=m#if for some odd reason, the deleted variables are referenced later, an out-of-bounds error is thrown
	n=len(H)
	k=len(constr.intervals)
	H=np.concatenate((H,np.zeros((n,k))),1)
	H=np.concatenate((H,np.zeros((k,n+k))),0)
	g=np.append(g,(k)*[0])
	
	for i in range(k):
		H[n+i,n+i]=1
	
	dx, H = linalg_solve(H, g, f, x, ev_constr)
	xi_full=np.zeros(m)
	OK=False
	keys=list(constr.intervals.keys())
	for r in range(50):		
		for j in range(k):
			key=keys[j]
			dx, H =kuhn_tucker_del(constr,key,j,n, H, g, x, f,dx,delmap, ev_constr)
		xi_full[idx]=dx[:n]
		OK=constr.within(x+xi_full,False)
		if OK: 
			break
		if r==k+3:
			#print('Unable to set constraints in computation calculation')
			break
	xi_full=np.zeros(m)
	xi_full[idx]=dx[:n]
	H_full = np.zeros((m,m))
	idx = idx.reshape((1,m))
	nz = np.nonzero(idx*idx.T)
	H_full[nz] = H[np.nonzero(np.ones((n,n)))]
	H_full[H_full==0] = H_orig[H_full==0]
	return xi_full, H_full


def kuhn_tucker_del(constr,key,j,n,H,g,x, f, dx,delmap, ev_constr,recalc=True):
	q=None
	c=constr.intervals[key]
	i=delmap[key]
	if not c.value is None:
		q=-(c.value-x[i])
	elif x[i]+dx[i]<c.min:
		q=-(c.min-x[i])
	elif x[i]+dx[i]>c.max:
		q=-(c.max-x[i])
	if q!=None:
		H[i,n+j]=1
		H[n+j,i]=1
		H[n+j,n+j]=0
		g[n+j]=q
		if recalc:
			dx, H = linalg_solve(H, g, f, x, ev_constr)
	return dx, H


def kuhn_tucker(c,i,j,n,H,g,x,f,dx, ev_constr,recalc=True):
	q=None
	if not c.value is None:
		q=-(c.value-x[i])
	elif x[i]+dx[i]<c.min:
		q=-(c.min-x[i])
	elif x[i]+dx[i]>c.max:
		q=-(c.max-x[i])
	if q!=None:
		H[i,n+j]=1
		H[n+j,i]=1
		H[n+j,n+j]=0
		g[n+j]=q
		if recalc:
			dx, H = linalg_solve(H, g, f, x, ev_constr)
	return dx, H


def kuhn_tucker2(c,i,j,n,H,g,x, f,dx,dx_init, ev_constr,recalc=True):
	if c.assco_ix is None:
		return kuhn_tucker(c,i,j,n,H,g,x, f, dx,ev_constr, recalc)
	q=None
	if not c.value is None:
		q=-(c.value-x[i])
	elif x[i]+dx[i]<c.min:
		q=-(c.min-x[i])
	elif x[i]+dx[i]>c.max:
		q=-(c.max-x[i])
	if q!=None:
		H[i,n+j]=1
		H[n+j,i]=1
		H[n+j,n+j]=0
		g[n+j]=q
		if recalc:
			dx, H = linalg_solve(H, g, f, x, ev_constr)
	return dx, H



	
def normalize(dx, x):
	x = np.abs(x)
	dx_norm=(x!=0)*dx/(x+(x<1e-100))
	dx_norm=(x<1e-2)*dx+(x>=1e-2)*dx_norm	
	return dx_norm	


	
def ev_non_sing(H, ev_constr):
	c, var_prop, ev, p = stat.var_decomposition(XXNorm=H)
	if not ev_constr:
		return ev,p
	c = c.flatten()
	p = p.real
	limit = 100
	v = np.sum(var_prop>0.5,1)<2
	keep = (v|(c<limit))
	ev = ev * keep
	return ev, p

def linalg_solve(H,g, f, x, ev_constr):
	
	ev, p = ev_non_sing(H, ev_constr)
	a = np.dot(p, p.T)
	b = np.dot(p.T, p)
	
	gp = np.dot(g, p)
	
	negev = (ev<0)+(ev==1.0)
	ev = ev*negev - ev*(negev==False)
	ev_1 = 1/(ev + (ev==0))
	
	gpL = gp*ev_1*(ev!=0)
	
	dx = -np.real(np.dot(gpL, p.T))
	
	H = np.dot(p, np.dot(np.diag(ev), p.T))

	
	return dx, H

def linalg_solve_old(H,g):
	try:
		return -np.linalg.solve(H,g).flatten(), H
	except np.linalg.LinAlgError as e:
		pass
	Hd = np.diag(H)
	Hd = Hd*(np.abs(Hd)>1e-30)
	Hd[Hd==0] = np.min(Hd[Hd<0])
	Hd = np.diag(Hd)
	H = H + Hd
	try:
		return -np.linalg.solve(H,g).flatten(), H
	except np.linalg.LinAlgError as e:
		pass		
	
	H = H - np.abs(Hd)
	try:
		return -np.linalg.solve(H,g).flatten(), H
	except np.linalg.LinAlgError as e:
		pass
	
	H =  - np.abs(Hd)
	try:
		return -np.linalg.solve(H,g).flatten(), H
	except np.linalg.LinAlgError as e:
		raise RuntimeError('Unable to fix singularity')		
	
	H = np.identity(len(Hd))
	return -np.linalg.solve(H,g).flatten(), H
		



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import sys
import os
import parallel
import traceback
import datetime
import pickle
import gc
import inspect
import time
from threading import Thread
import subprocess
from queue import Queue
from queue import Empty
import callback


class Slave:
	def __init__(self, t, s_id, path, print_to_file, n_nodes):
		self.t = t
		self.id = s_id
		self.n_nodes = n_nodes
		self.d = dict()
		self.d['slave_id'] = s_id
		self.processes = Processes(self)
		while 1:
			self.wait_for_orders()

		
	def wait_for_orders(self):
	
		(msg,obj) = self.t.receive()
		response = None
		if msg == 'kill':
			sys.exit()
			response = True
			
		elif msg == 'callback':
			name, d = obj
			if name in self.processes:
				response = dict(self.processes[name].transact(d))
				response['slave_id'] = self.id
			else:
				response = {'empty':True}		
				
		elif msg == 'check_state':
			response = self.processes.countalive(obj)
		elif msg == 'collect':
			response = self.processes.collect(obj)		
		else:#starting a new procedure
			self.processes[msg] = obj
		self.t.send(response)	
		gc.collect()
	


class Processes(dict):
	
	def __init__(self,slave):
		self.d = slave.d
		self.parent = slave
		
	def __setitem__(self, name, task):
		super().__setitem__(name, Process(name, task, self))
		
	def countalive(self, name = None):
		if not name is None:
			if name in self:
				return self[name].thread.is_alive()
			return 0
		count = 0
		for k in self:
			if hasattr(self[k],'thread'):
				if self[k].thread.is_alive():
					count+=1
		return count
	
	def collect(self, name):
		if not name in self:
			return None, None
			#raise RuntimeError(f"The process '{name}' does not exist in node {self.parent.id}")
		d, r = self[name].collect()
		for k in d:
			self.d[k] = d[k]
		return r
		
		

	
class Process:
	def __init__(self, name, task, parent):			
		self.name = name
		self.parent = parent
		self.task = task
		self.d = dict(parent.d)
		self.d['inbox'] = {}
		self.d['outbox'] = {}
		self.threadq = Queue()
		self.inbox = self.d['outbox']
		self.outbox = self.d['inbox']
		self.result = {}, None
		
		if name == 'transfer dictionary':
			t0=time.time()
			self.get_dict(task)
		else:
			if name == 'debug':
				execute_task(self.d, task, self.threadq)
			else:
				prtask = task.replace('\n', ' ')
				#print(f"id inbox: {id(self.inbox)}, id outbox:{id(self.outbox)}, task:{prtask}")
				self.thread = Thread(target=execute_task, args=(self.d, task, self.threadq, self.name))
				self.thread.start()	
		a=0
		
	def get_dict(self, fname):	#Blocking	
		try:
			f = open(fname,'rb')
			d = None
			d =  pickle.load(f)
			f.close()	
			for k in d:
				self.parent.d[k] = d[k]
	
		except Exception as e:
			traceback.print_exc(file = sys.stdout)
			raise e
		
	def transact(self, incoming):
		#print(f'reading from {id(self.inbox)}')
		if not self.thread.is_alive():
			return dict(self.inbox)
		for k in incoming:
			self.outbox[k] = incoming[k]	
		return dict(self.inbox)

	
	def collect(self):
		if self.threadq.empty():
			return self.result
		if not self.threadq.get()=='start':
			raise RuntimeError("Expected a start signal")
		d, r = self.threadq.get()
		self.result = d, r
		return d, r
		
		
		

def execute_task(d, task, queue,name):#None blocking 
	queue.put('start')
	d = dict(d)
	try:
		r = execute_task_(d, task)			
	except Exception as e:
		traceback.print_exc(file = sys.stdout)
		raise e
	queue.put((d,r))

	
def execute_task_(d, task):
	try:#If a returned value is desired, the task must be a function call
		return eval(task, globals(), d)
	except SyntaxError as e:#else, it is assumed that the script only generates local variables for later use
		exec(task, globals(), d)

						   


def run(transact, print_to_file = True):
	if not print_to_file:
		msg, (s_id, path, n_nodes), fname = handshake(transact, print_to_file)
		#Wait for instructions:
		Slave(transact, s_id, path, print_to_file, n_nodes)			
		return
	try: 
		msg, (s_id, path, n_nodes), fname = handshake(transact, print_to_file)
		f = open(fname,'w', 1)
		f.close()
		#Wait for instructions:		
		fstdout = os.path.join(path,f'slaves/{s_id}.txt')
		if print_to_file:
			sys.stdout = open(fstdout,'w',1)	
		Slave(transact, s_id, path, print_to_file, n_nodes)
	except Exception as e:
		print(f'error was: {e}')
		try:
			traceback.print_exc(file=sys.stdout)
		except Exception as e:
			print(e)
		f.write('SID: %s      TIME:%s \n' %(s_id,datetime.datetime.now()))
		traceback.print_exc(file = f)
		f.flush()
		f.close()
		raise e
	
def handshake(transact, print_to_file):
	#Handshake:
	transact.send(os.getpid())
	path='.'
	msg, (s_id, path, n_nodes)=transact.receive()
	transact.set('slave', n_nodes, path, s_id)
	#error handling
	fname = os.path.join(path,'slave_errors.txt')	
	return msg, (s_id, path, n_nodes), fname
			

def write(f,txt):
	f = open(f, 'w', 1)
	f.write(str(txt))
	f.flush()
	f.close()	


#!/usr/bin/env python
# -*- coding: utf-8 -*-

#used for debugging
import numpy as np
import calculus_functions as cf
import time
import os
import loglikelihood as lgl



def hess_debug(ll,panel,g,d):
	"""Calculate the Hessian nummerically, using the analytical gradient. For debugging. Assumes correct and debugged gradient"""
	x=ll.args.args_v
	n=len(x)
	dx=np.identity(n)*d
	H=np.zeros((n,n))
	ll0=lgl.LL(x,panel)
	f0=g.get(ll0)
	for i in range(n):
		ll=lgl.LL(x+dx[i],panel)
		if not ll is None:
			f1=g.get(ll)
			H[i]=(f1-f0)/d

			
	return H

def grad_debug(ll,panel,d):
	"""Calcualtes the gradient numerically. For debugging"""
	x=ll.args.args_v
	n=len(x)
	dx=np.abs(x.reshape(n,1))*d
	dx=dx+(dx==0)*d
	dx=np.identity(n)*dx

	g=np.zeros(n)
	f0=lgl.LL(x,panel)
	for i in range(n):
		for j in range(5):
			dxi=dx[i]*(0.5**j)
			f1=lgl.LL(x+dxi,panel)
			if not f1 is None:
				if not f1.LL is None:
					g[i]=(f1.LL-f0.LL)/dxi[i]
					break
	return g


	
def grad_debug_detail(f0,panel,d,llname,varname1,pos1=0):
	args1=lgl.copy_array_dict(f0.args.args_d)
	args1[varname1][pos1]+=d
	
	f0=lgl.LL(f0.args.args_d, panel)
	f1=lgl.LL(args1, panel)

	if type(llname)==list or type(llname)==tuple:
		ddL=(f1.__dict__[llname[0]].__dict__[llname[1]]-f0.__dict__[llname[0]].__dict__[llname[1]])/d
	else:
		ddL=(f1.__dict__[llname]-f0.__dict__[llname])/d
	return ddL

def test_c_armas(u_RE, var_ARMA, e_RE, panel, ll):
	var_ARMA2 = panel.arma_dot.dot(ll.GAR_1MA,ll.h_val,ll)
	e_RE2 = panel.arma_dot.dot(ll.AMA_1AR,u_RE,ll)	
	print(f"Testsums arma: c:{np.sum(var_ARMA**2)}, py:{np.sum(var_ARMA2**2)}")
	print(f"Testsums e: c:{np.sum(e_RE**2)}, py:{np.sum(e_RE2**2)}")

	
	
def hess_debug_detail(f0,panel,d,llname,varname1,varname2,pos1=0,pos2=0):
	args1=lgl.copy_array_dict(f0.args.args_d)
	args2=lgl.copy_array_dict(f0.args.args_d)
	args3=lgl.copy_array_dict(f0.args.args_d)
	args1[varname1][pos1]+=d
	args2[varname2][pos2]+=d	
	args3[varname1][pos1]+=d
	args3[varname2][pos2]+=d
	f1=lgl.LL(args1, panel)
	f2=lgl.LL(args2, panel)
	f3=lgl.LL(args3, panel)
	if type(llname)==list:
		ddL=(f3.__dict__[llname[0]].__dict__[llname[1]]-f2.__dict__[llname[0]].__dict__[llname[1]]
		     -f1.__dict__[llname[0]].__dict__[llname[1]]+f0.__dict__[llname[0]].__dict__[llname[1]])/(d**2)
	else:
		ddL=(f3.__dict__[llname]-f2.__dict__[llname]-f1.__dict__[llname]+f0.__dict__[llname])/(d**2)
	return ddL
	


def LL_calc(self,panel):
	panel=self.panel
	X=panel.XIV
	matrices=set_garch_arch(panel,self.args.args_d)
	if matrices is None:
		return None		
	
	AMA_1,AMA_1AR,GAR_1,GAR_1MA=matrices
	(N,T,k)=X.shape
	#Idea for IV: calculate Z*u throughout. Mazimize total sum of LL. 
	u = panel.Y-cf.dot(X,self.args.args_d['beta'])
	e = cf.dot(AMA_1AR,u)
	e_RE = (e+self.re_obj_i.RE(e, panel)+self.re_obj_t.RE(e, panel))*panel.included[3]
	
	e_REsq =(e_RE**2+(e_RE==0)*1e-18) 
	grp = self.variance_RE(panel,e_REsq)#experimental
	
	W_omega = cf.dot(panel.W_a, self.args.args_d['omega'])

	lnv_ARMA = self.garch(panel, GAR_1MA, e_RE)
	
	lnv = W_omega+lnv_ARMA# 'N x T x k' * 'k x 1' -> 'N x T x 1'
	lnv+=grp
	self.dlnv_pos=(lnv<100)*(lnv>-100)
	lnv = np.maximum(np.minimum(lnv,100),-100)
	v = np.exp(lnv)*panel.a[3]
	v_inv = np.exp(-lnv)*panel.a[3]

	LL = self.LL_const-0.5*(lnv+(e_REsq)*v_inv)

	self.tobit(panel,LL)
	LL=np.sum(LL*panel.included[3])
			
	self.add_variables(matrices, u, e, lnv_ARMA, lnv, v, W_omega, grp,e_RE,e_REsq,v_inv)
	if abs(LL)>1e+100: 
		return None				
	return LL


def save_reg_data(ll, panel, fname = 'repr.csv'):
	#saves data neccessary to reproduce 
	N,T,k = panel.X.shape
	a = np.concatenate((panel.X, panel.Y, panel.W, ll.e, ll.var), 2)
	a = a.reshape((T, a.shape[2]))
	coefs = np.zeros((T,1))
	coef_arr = ll.args.args_v
	coefs[:len(coef_arr),0] = coef_arr
	a = np.concatenate((a, 
						ll.LL_full[0].reshape((T,1)),
						ll.AMA_1AR[0].reshape((T,1))[::-1],
						ll.GAR_1MA[0].reshape((T,1))[::-1],
						coefs
						),1)
	np.savetxt(fname, a, fmt='%s', delimiter=';')	


#!/usr/bin/env python
# -*- coding: utf-8 -*-


import numpy as np
import stat_functions as stat
import stat_dist
import os
from matplotlib import pyplot  as plt



class ProcessCharts():
	def __init__(self,panel):
		self.panel=panel
		self.ll=None	
		self.subplot=plt.subplots(1,figsize=(4,2.5),dpi=75)
		self.chart_list=[
			['histogram',self.histogram],
			['correlogram',self.correlogram],
			['correlogram_variance',self.correlogram_variance]
		]

	def save_all(self,ll):
		self.ll=ll
		if self.ll is None:
			return
		for name,chart in self.chart_list:
			chart(f'img/{name}.png')				

	def histogram(self,f):
		N,T,k=self.panel.X.shape
		fgr,axs = self.subplot
		n=self.ll.e_norm_centered.shape[2]
		e=self.ll.e_norm_centered[self.panel.included[2]].flatten()
		N=e.shape[0]
		e=e.reshape((N,1))

		grid_range=4
		grid_step=0.05	
		h,grid=histogram(e,grid_range,grid_step)
		norm=stat_dist.norm(grid, cdf = False)*grid_step	

		axs.bar(grid,h,color='grey', width=0.025,label='histogram')
		axs.plot(grid,norm,'green',label='normal distribution')
		axs.legend(prop={'size': 6})
		name='Histogram - frequency'
		axs.set_title(name)
		self.save(f)

	def correlogram(self,f):
		fgr,axs=self.subplot
		lags=20
		rho=stat.correlogram(self.panel, self.ll.e_norm_centered,lags)
		x=np.arange(lags+1)
		axs.bar(x,rho,color='grey', width=0.5,label='correlogram')
		name='Correlogram - residuals'
		axs.set_title(name)
		self.save(f)

	def correlogram_variance(self,f):
		N,T,k=self.panel.X.shape
		fgr,axs=self.subplot
		lags=20
		e2=self.ll.e_norm_centered**2
		e2=(e2-self.panel.mean(e2))*self.panel.included[3]
		rho=stat.correlogram(self.panel, e2,lags)
		x=np.arange(lags+1)
		axs.bar(x,rho,color='grey', width=0.5,label='correlogram')
		name='Correlogram - squared residuals'
		axs.set_title(name)
		self.save(f)
		
	def save(self,save_file):
		fgr,axs=self.subplot
		fgr.savefig(save_file)
		axs.clear()	

def histogram(x,grid_range,grid_step):
	N,k=x.shape
	grid_n=int(2*grid_range/grid_step)
	grid=np.array([i*grid_step-grid_range for i in range(grid_n)]).reshape((1,grid_n))
	ones=np.ones((N,1))
	x_u=np.concatenate((ones,x>=grid),1)
	x_l=np.concatenate((x<grid,ones),1)
	grid=np.concatenate((grid.flatten(),[grid[0,-1]+grid_step]))
	histogram=np.sum((x_u*x_l),0)
	if int(np.sum(histogram))!=N:
		raise RuntimeError('Error in histogram calculation')
	return histogram/N,grid





#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import sys
import os
sys.path.append(__file__.replace("__init__.py",''))
import time
import matplotlib
import parallel


PARALLEL_LAYER1 = False
PARALLEL_LAYER2 = True
CALLBACK_ACTIVE = True
DIRECT = False
if DIRECT:
	PARALLEL_LAYER2 = False
	
	
N_NODES = 10

t0=time.time()

path = os.getcwd().replace('\\', '/')
subpath = os.path.join(path,'mp').replace('\\', '/')

mp = parallel.Parallel(1, path, PARALLEL_LAYER1, CALLBACK_ACTIVE, DIRECT)


mp.exec(["import maximize\n"
		"import parallel as parallel\n"
		f"mp = parallel.Parallel({N_NODES},'{subpath}', {PARALLEL_LAYER2}, {CALLBACK_ACTIVE}, {DIRECT}, 1)\n" 
		"mp.exec('import loglikelihood as logl\\n'\n"
		"'import maximize', 'init')\n"
		"outbox['mp'] = mp\n"], 'init')

print(f"parallel: {time.time()-t0}")


import pandas as pd


import output
import main
import options as opt_module
import inspect
import loaddata
import psutil
import signal
from threading import Thread
import traceback


def execute(model_string,dataframe, ID=None,T=None,HF=None,instruments=None, console_output=True):

	"""Maximizes the likelihood of an ARIMA/GARCH model with random/fixed effects (RE/FE)\n
	model_string: a string on the form 'Y ~ X1 + X2 + X3\n
	dataframe: a dataframe consisting of variables with the names usd in model_string, ID, T, HF and instruments\n
	ID: The group identifier\n
	T: the time identifier\n
	HF: list with names of heteroskedasticity factors (additional regressors in GARCH)\n
	instruments: list with names of instruments
	console_output: if True, GUI output is turned off (GUI output is experimental)
	"""
	
	window=main.identify_global(inspect.stack()[1][0].f_globals,'window')
	exe_tab=main.identify_global(inspect.stack()[1][0].f_globals,'exe_tab')
	try:
		r=main.execute(model_string,dataframe,ID, T,HF,options,window,exe_tab,instruments, console_output, mp, PARALLEL_LAYER2)
	except KeyboardInterrupt as e:
		print('interrrupt')
		kill_orpahns()
	except Exception as e:
		print('exept')
		traceback.print_exc(file=sys.stdout)
		print('except2')
		kill_orpahns()
		raise e
	return r


def kill_orpahns():
	try:
		pids = mp.callback('init')[0]['mp'].pids
	except:
		pids = []
	pids = [int(i) for i in pids]
	kill(pids)
	kill([int(mp.pids[0])])
	kill([os.getpid()])
	
	

def kill(pids):
	for proc in psutil.process_iter():
		pid = proc.pid
		if pid in pids:
			try:
				os.kill(proc.pid, signal.SIGTERM)
				print(f"killed pid {pid}")
			except psutil.NoSuchProcess:
				pass


def load_json(fname):

	if False:#detects previously loaded dataset in the environment
		dataframe=main.indentify_dataset(globals(),fname)
		if (not dataframe==False) and (not dataframe is None):
			return dataframe	
	try:
		dataframe=main.loaddata.load_json(fname)
	except FileNotFoundError:
		raise RuntimeError("File %s not found" %(fname))
	return dataframe


def load(fname,sep=None):

	"""Loads data from file <fname>, asuming column separator <sep>.\n
	Returns a dataframe (a dictionary of numpy column matrices).\n
	If sep is not supplied, the method will attemt to find it."""
	if False:#detects previously loaded dataset in the environment
		dataframe=main.indentify_dataset(globals(),fname)
		if (not dataframe==False) and (not dataframe is None):
			return dataframe	
	try:
		dataframe=main.loaddata.load(fname,sep)
	except FileNotFoundError:
		raise RuntimeError("File %s not found" %(fname))
	return dataframe

def load_SQL(conn,sql_string):

	"""Loads data from an SQL server, using sql_string as query"""
	if False:#detects previously loaded dataset in the environment
		dataframe=main.indentify_dataset(globals(),sql_string)
		if (not dataframe==False) and (not dataframe is None):
			return dataframe
	dataframe=main.loaddata.load_SQL(sql_string,conn)
	#except RuntimeError as e:
	#	raise e
	return dataframe
		
	
options=opt_module.regression_options()
preferences=opt_module.application_preferences()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import time

try:
	import cfunctions as c
except:
	c=None


def dd_func_lags_mult(panel,ll,g,AMAL,vname1,vname2,transpose=False, u_gradient=False):
	de2_zeta_xi_RE,de2_zeta_xi 	= dd_func_lags_mult_arima(panel,ll,g,AMAL,vname1,vname2,transpose, u_gradient)
	dd_re_variance 							= dd_func_re_variance(panel,ll,g,vname1,vname2,de2_zeta_xi_RE,u_gradient)
	d2LL_d2var_zeta_xi,d2LL_d2e_zeta_xi_RE 	= dd_func_garch(panel,ll,g,vname1,vname2,de2_zeta_xi_RE,de2_zeta_xi,dd_re_variance,u_gradient)
	
	return d2LL_d2var_zeta_xi,d2LL_d2e_zeta_xi_RE

def dd_func_lags_mult_arima(panel,ll,g,AMAL,vname1,vname2,transpose, u_gradient):
	#de_xi is "N x T x m", de_zeta is "N x T x k" and L is "T x T"
	de_xi=g.__dict__['de_'+vname1]
	de_zeta=g.__dict__['de_'+vname2]
	de2_zeta_xi_RE=None
	if de_xi is None or de_zeta is None:
		return None,None
	(N,T,m)=de_xi.shape
	(N,T,k)=de_zeta.shape	
	#ARIMA:
	if not AMAL is None:
		if u_gradient:
			de2_zeta_xi=-panel.arma_dot.dot(AMAL,g.X_RE,ll)#"T x N x s x m #for error beta-rho covariance, the u gradient must be used	
		else:
			de2_zeta_xi=panel.arma_dot.dot(AMAL,de_zeta,ll)#"T x N x s x m
		if transpose:#only happens if lags==k
			de2_zeta_xi=de2_zeta_xi+np.swapaxes(de2_zeta_xi,2,3)#adds the transpose
		de2_zeta_xi=de2_zeta_xi*panel.included[4]
	else:
		de2_zeta_xi=None
	de2_zeta_xi_RE=de2_zeta_xi
	
	return de2_zeta_xi_RE,de2_zeta_xi


def dd_func_re_variance(panel,ll,g,vname1,vname2,de2_zeta_xi_RE,u_gradient):
	if panel.N<=1 or panel.options.fixed_random_group_eff.value==0:
		return None
	#voaltility RE:

	de_xi_RE=g.__dict__['de_'+vname1+'_RE']
	de_zeta_RE=g.__dict__['de_'+vname2+'_RE']
	if de_xi_RE is None or de_zeta_RE is None:
		return None
	(N,T,m)=de_xi_RE.shape
	(N,T,k)=de_zeta_RE.shape		
	de_xi_RE_r=de_xi_RE.reshape((N,T,m,1))
	de_zeta_RE_r=de_zeta_RE.reshape((N,T,1,k))	
	incl=panel.included[4]
	
	dvRE_xi=g.__dict__['dvRE_'+vname1].reshape((N,T,m,1))
	dvRE_zeta=g.__dict__['dvRE_'+vname2].reshape((N,T,1,k))			
	
	if not de2_zeta_xi_RE is None:
		d_xi_input=g.__dict__['d_'+vname1+'_input']
		d_zeta_input=g.__dict__['d_'+vname2+'_input']		
		try:
			dd_e_RE_sq=(2*de_xi_RE_r*de_zeta_RE_r+2*ll.e_RE.reshape(N,T,1,1)*de2_zeta_xi_RE)*incl
		except:
			dd_e_RE_sq=(2*de_xi_RE_r*de_zeta_RE_r+2*ll.e_RE.reshape(N,T,1,1)*de2_zeta_xi_RE)*incl
		ddmeane2= panel.mean(dd_e_RE_sq,(0,1))*incl
		dd_input=(dd_e_RE_sq-ddmeane2)*incl

		ddvRE_d_xi_zeta=ddmeane2-add((ll.re_obj_i_v.ddRE(dd_input,d_xi_input,d_zeta_input,ll.varRE_input,vname1,vname2,panel),
									  ll.re_obj_t_v.ddRE(dd_input,d_xi_input,d_zeta_input,ll.varRE_input,vname1,vname2,panel)),True)
	else:
		ddvRE_d_xi_zeta=None
		
	dvarRE,		ddvarRE		=	ll.dvarRE.reshape((N,T,1,1)),		ll.ddvarRE.reshape((N,T,1,1))
	dd_re_variance	=	add((prod((dvarRE,ddvRE_d_xi_zeta))	,	ddvarRE*dvRE_xi*dvRE_zeta))
	return dd_re_variance


def dd_func_garch(panel,ll,g,vname1,vname2,de2_zeta_xi_RE,de2_zeta_xi,dd_re_variance,u_gradient):
	#GARCH: 
	incl=panel.included[4]
	if (g.__dict__['de_'+vname1] is None) or (g.__dict__['de_'+vname2] is None):
		return None, None	
	(N,T,m)=g.__dict__['de_'+vname1].shape
	(N,T,k)=g.__dict__['de_'+vname2].shape
	DLL_e=g.DLL_e.reshape(N,T,1,1)
	
	d2LL_d2e_zeta_xi_RE=None
	if not de2_zeta_xi_RE is None:	
		d2LL_d2e_zeta_xi_RE = de2_zeta_xi_RE * DLL_e	
		d2LL_d2e_zeta_xi_RE = np.sum(np.sum(d2LL_d2e_zeta_xi_RE*incl,0),0)
		
	RE_suffix=''

	de_xi_RE=g.__dict__['de_'+vname1+RE_suffix]
	de_zeta_RE=g.__dict__['de_'+vname2+RE_suffix]	

	
	dLL_var=g.dLL_var.reshape(N,T,1,1)
	if de2_zeta_xi_RE is None:
		de2_zeta_xi_RE=0
	d2var_zeta_xi=None
	d2LL_d2var_zeta_xi=None
	d_omega_e=None
	d2var_zeta_xi_h=None
	if panel.pqdkm[4]>0:
		if u_gradient:
			de_zeta_RE=g.__dict__['de_'+vname2+RE_suffix]
		h_e_de2_zeta_xi =  ll.h_e_val.reshape(N,T,1,1)  * de2_zeta_xi_RE
		h_2e_dezeta_dexi = ll.h_2e_val.reshape(N,T,1,1) * de_xi_RE.reshape((N,T,m,1)) * de_zeta_RE.reshape((N,T,1,k))

		d2var_zeta_xi_h = (h_e_de2_zeta_xi + h_2e_dezeta_dexi)
		
		d2var_zeta_xi_h = panel.arma_dot.dot(ll.GAR_1MA, d2var_zeta_xi_h,ll)
		
	d2var_zeta_xi = add((d2var_zeta_xi_h,  dd_re_variance,d_omega_e), True)
	if not d2var_zeta_xi is None:
		d2LL_d2var_zeta_xi = np.sum(prod((d2var_zeta_xi,dLL_var*incl),True),(0,1))
	
	return d2LL_d2var_zeta_xi,d2LL_d2e_zeta_xi_RE

def dd_func_lags(panel,ll,L,d,dLL,transpose=False):
	#d is "N x T x m" and L is "k x T x T"
	if panel.pqdkm[4]==0:
		return None
	if d is None:
		return None		
	(N,T,m)=d.shape
	if L is None:
		x=0
	elif len(L)==0:
		return None
	elif type(L)==tuple:#elif len(L.shape)==3:
		x=panel.arma_dot.dot(L,d,ll)#"T x N x k x m"
		if x is None:
			return None
	elif len(L.shape)==2:
		x=dot(L,d).reshape(N,T,1,m)
	dLL=dLL.reshape((N,T,1,1))
	return np.sum(np.sum(dLL*x,1),0)#and sum it	


def add(iterable,ignore=False):
	"""Sums iterable. If ignore=True all elements except those that are None are added. If ignore=False, None is returned if any element is None. """
	x=None
	for i in range(len(iterable)):
		if not iterable[i] is None:
			if x is None:
				x=iterable[i]
			else:
				x=x+iterable[i]
		else:
			if not ignore:
				return None
	return x

def prod(iterable,ignore=False):
	"""Takes the product sum of iterable. If ignore=True all elements except those that are None are multiplied. 
	If ignore=False, None is returned if any element is None. """
	x=None
	for i in iterable:
		if not i is None:
			if x is None:
				x=i
			else:
				x=x*i
		else:
			if not ignore:
				return None
	return x

def sumNT(nparray):
	if nparray is None:
		return None
	s=nparray.shape
	if len(s)<3:
		raise RuntimeError("Not enough dimensions")
	
	return np.sum(nparray.reshape(list(s)+[1]),(0,1))
	

def concat_matrix(block_matrix):
	m=[]
	for i in range(len(block_matrix)):
		r=block_matrix[i]
		C=[]
		for j in range(len(r)):
			if not r[j] is None:
				C.append(r[j])
		if len(C):
			m.append(np.concatenate(C,1))
	m=np.concatenate(m,0)
	return m

def concat_marray(matrix_array):
	arr=[]
	for i in matrix_array:
		if not i is None:
			arr.append(i)
	arr=np.concatenate(arr,2)
	return arr
		
def dd_func(d2LL_de2,d2LL_dln_de,d2LL_dln2,de_dh,de_dg,dln_dh,dln_dg,dLL_de2_dh_dg,dLL_dln2_dh_dg):
	a=[]
	a.append(dd_func_mult(de_dh,d2LL_de2,de_dg))

	a.append(dd_func_mult(de_dh,d2LL_dln_de,dln_dg))
	a.append(dd_func_mult(dln_dh,d2LL_dln_de,de_dg))

	a.append(dd_func_mult(dln_dh,d2LL_dln2,dln_dg))

	a.append(dLL_de2_dh_dg)
	a.append(dLL_dln2_dh_dg)
	return add(a,True)

def dd_func_mult(d0,mult,d1):
	#d0 is N x T x k and d1 is N x T x m
	if d0 is None or d1 is None or mult is None:
		return None
	(N,T,k)=d0.shape
	(N,T,m)=d1.shape
	if np.any(np.isnan(d0)) or np.any(np.isnan(d1)):
		x=np.empty((k,m))
		x[:]=np.nan
		return x
	d0=d0*mult
	d0=np.reshape(d0,(N,T,k,1))
	d1=np.reshape(d1,(N,T,1,m))
	try:
		x=np.sum(np.sum(d0*d1,0),0)#->k x m 
	except RuntimeWarning as e:
		if e.args[0]=='overflow encountered in multiply':
			d0=np.minimum(np.maximum(d0,-1e+100),1e+100)
			d1=np.minimum(np.maximum(d1,-1e+100),1e+100)
			x=np.sum(np.sum(d0*d1,0),0)#->k x m 
		else:
			raise RuntimeWarning(e)
	return x


def dot(a,b,reduce_dims=True):
	"""Matrix multiplication. Returns the dot product of a*b where either a or be or both to be
	arrays of matrices. Faster than mmult, less general and only used for special purpose.
	Todo: generalize and merge"""


	if len(a.shape)==3 and len(b.shape)==2:
		x = np.array([np.dot(a[i],b) for i in range(a.shape[0])])
	elif len(a.shape)==3 and len(b.shape)==3:
		x = np.sum([np.dot(a[i].T,b[i]) for i in range(a.shape[0])],0)
	elif len(a.shape)==2 and len(b.shape)==3:
		x = np.array([np.dot(a,b[i]) for i in range(b.shape[0])])
	return x


class arma_dot_obj:
	def __init__(self):
		pass
		
	def dotroll(self,aband,k,sign,b,ll):
		x = sign*self.fast_dot(aband, b)
		w=[]
		for i in range(k):
			w.append(np.roll(np.array(x),i+1,1))
			w[i][:,:i+1]=0
		x=np.array(w)
		x=np.moveaxis(x,0,2)
		return x
			
			
	def fast_dot(self, a, b):
		a_, name = a
		n = get_n(a_)
		if n is None:
			n = len(a_)
			
		r = a_[0]*b
		for i in range(1,n):
			r[:,i:] += a_[i]*b[:,:-i]
			
		return r

		
	def dot(self,a,b,ll):
		if len(a)>2:#then this is a proper matrix
			(aband,k,sgn)=a
			if k==0:
				return None
			return self.dotroll(aband, k, sgn, b, ll)
		x = self.fast_dot(a, b)
		return x


def get_n(a):
	minval = 0
	a_1 = np.abs(a[1:])
	max_a = np.max(a_1)
	if np.min(np.abs(a_1)) >= minval:
		return None
	if max_a == 0:
		return 1		
	else:
		nz = np.nonzero(a_1/max_a < minval)[0]
		if len(nz)>0:
			return nz[0]+1
		else:
			return None


#!/usr/bin/env python
# -*- coding: utf-8 -*-

#contains the log likelihood object
import sys
from pathlib import Path
import os
import numpy.ctypeslib as npct
import ctypes as ct
p = os.path.join(Path(__file__).parent.absolute(),'cfunctions')
cfunct = npct.load_library('ctypes.dll',p)

import numpy as np
import calculus_ll as cll
import calculus_functions as cf
import random_effects as re
import traceback
import model_parser
import time
import stat_dist
import stat_functions


CDPT = ct.POINTER(ct.c_double) 
CIPT = ct.POINTER(ct.c_uint) 

	

class LL:
	"""Calculates the log likelihood given arguments arg (either in dictonary or array form), and creates an object
	that store dynamic variables that depend on the \n
	If args is a dictionary, the ARMA-GARCH orders are 
	determined from the dictionary. If args is a vector, the ARMA-GARCH order needs to be consistent
	with the  panel object
	"""
	def __init__(self,args,panel,constraints=None,print_err=False):
		self.err_msg=''
		self.errmsg_h=''

		#checking settings. If the FE/RE is done on the data before LL
		gfre=panel.options.fixed_random_group_eff.value
		tfre=panel.options.fixed_random_time_eff.value
		vfre=panel.options.fixed_random_variance_eff.value
		
		self.re_obj_i=re.re_obj(panel,True,panel.T_i,panel.T_i,gfre)
		self.re_obj_t=re.re_obj(panel,False,panel.date_count_mtrx,panel.date_count,tfre)
		self.re_obj_i_v=re.re_obj(panel,True,panel.T_i,panel.T_i,gfre*vfre)
		self.re_obj_t_v=re.re_obj(panel,False,panel.date_count_mtrx,panel.date_count,tfre*vfre)
		
		self.args=panel.args.create_args(args,panel,constraints)
		self.h_err=""
		self.LL=None
		#self.LL=self.LL_calc(panel)
		try:
			self.LL=self.LL_calc(panel)
			if np.isnan(self.LL):
				self.LL=None						
		except Exception as e:
			if print_err:
				traceback.print_exc()
				print(self.errmsg_h)
		
		


	def LL_calc(self,panel):
		X=panel.XIV
		incl = panel.included[3]

		
		#Idea for IV: calculate Z*u throughout. Mazimize total sum of LL. 
		u = panel.Y-cf.dot(X,self.args.args_d['beta'])
		u_RE = (u+self.re_obj_i.RE(u, panel)+self.re_obj_t.RE(u, panel))*incl
		
		matrices=self.arma_calc(panel, u_RE)
		if matrices is None:
			return None		
		AMA_1,AMA_1AR,GAR_1,GAR_1MA, e_RE, var_ARMA, h=matrices	
			

		if panel.options.EGARCH.value==0:
			e_REsq =(e_RE**2+(e_RE==0)*1e-18) 
			nd =1
			self.h_val, self.h_e_val, self.h_2e_val = e_RE**2*incl, nd*2*e_RE*incl, nd*2*incl
			self.h_z_val, self.h_2z_val,  self.h_ez_val = None,None,None	#possibility of additional parameter in sq res function		
		else:
			minesq = 1e-100
			e_REsq =np.maximum(e_RE**2,minesq)
			nd = e_RE**2<minesq			
			self.h_val, self.h_e_val, self.h_2e_val = np.log(e_REsq)*incl, nd*2*np.sign(e_RE)/(e_REsq)**0.5, -2*nd*incl/(e_REsq)
			self.h_z_val, self.h_2z_val,  self.h_ez_val = None,None,None	#possibility of additional parameter in sq res function		
		
		W_omega = cf.dot(panel.W_a, self.args.args_d['omega'])
		
		
		#import debug
		#debug.test_c_armas(u_RE, var_ARMA, e_RE, panel, self)
		
		
		var = W_omega+var_ARMA

		LL_full,v,v_inv,self.dvar_pos=cll.LL(panel,var,e_REsq, e_RE)
		self.tobit(panel,LL_full)
		LL=np.sum(LL_full*incl)
		self.LL_all=np.sum(LL_full)
		self.add_variables(panel,matrices, u, u_RE, var_ARMA, var, v, W_omega,e_RE,e_REsq,v_inv,LL_full)
		if abs(LL)>1e+100: 
			return None				
		return LL
		
	def add_variables(self,panel,matrices,u, u_RE,var_ARMA,var,v,W_omega,e_RE,e_REsq,v_inv,LL_full):
		self.v_inv05=v_inv**0.5
		self.e_norm=e_RE*self.v_inv05	
		self.e_norm_centered=(self.e_norm-panel.mean(self.e_norm))*panel.included[3]
		self.u, self.u_RE, self.var_ARMA        = u,  u_RE,   var_ARMA
		self.var,  self.v,    self.LL_full = var,       v,    LL_full
		self.W_omega=W_omega
		self.e_RE=e_RE
		self.e_REsq=e_REsq
		self.v_inv=v_inv


	
	def tobit(self,panel,LL):
		if sum(panel.tobit_active)==0:
			return
		g=[1,-1]
		self.F=[None,None]	
		for i in [0,1]:
			if panel.tobit_active[i]:
				I=panel.tobit_I[i]
				self.F[i]= stat_dist.norm(g[i]*self.e_norm[I])
				LL[I]=np.log(self.F[i])
	
	
	def variance_RE(self,panel,e_REsq):
		"""Calculates random/fixed effects for variance."""
		#not in use, expermental. Should be applied to normalize before ARIMA/GARCH
		self.vRE,self.varRE,self.dvarRE=panel.zeros[3],panel.zeros[3],panel.zeros[3]
		self.ddvarRE,self.dvarRE_mu,self.ddvarRE_mu_vRE=panel.zeros[3],None,None
		self.varRE_input, self.ddvarRE_input, self.dvarRE_input = None, None, None
		if panel.options.fixed_random_variance_eff.value==0:
			return panel.zeros[3]
		if panel.N==0:
			return None

		meane2=panel.mean(e_REsq)
		self.varRE_input=(e_REsq-meane2)*panel.included[3]

		mine2=0
		mu=panel.options.variance_RE_norm.value
		self.vRE_i=self.re_obj_i_v.RE(self.varRE_input, panel)
		self.vRE_t=self.re_obj_t_v.RE(self.varRE_input, panel)
		self.meane2=meane2
		vRE=meane2*panel.included[3]-self.vRE_i-self.vRE_t
		self.vRE=vRE
		small=vRE<=mine2
		big=small==False
		vREbig=vRE[big]
		vREsmall=vRE[small]

		varREbig=np.log(vREbig+mu)
		varREsmall=(np.log(mine2+mu)+((vREsmall-mine2)/(mine2+mu)))
		varRE,dvarRE,ddvarRE=np.zeros(vRE.shape),np.zeros(vRE.shape),np.zeros(vRE.shape)
		
		varRE[big]=varREbig
		varRE[small]=varREsmall
		self.varRE=varRE*panel.included[3]

		dvarRE[big]=1/(vREbig+mu)
		dvarRE[small]=1/(mine2+mu)
		self.dvarRE=dvarRE*panel.included[3]
		
		ddvarRE[big]=-1/(vREbig+mu)**2
		self.ddvarRE=ddvarRE*panel.included[3]
	
		return self.varRE
		


	def standardize(self,panel,reverse_difference=False):
		"""Adds X and Y and error terms after ARIMA-E-GARCH transformation and random effects to self. 
		If reverse_difference and the ARIMA difference term d>0, the standardized variables are converted to
		the original undifferenced order. This may be usefull if the predicted values should be used in another 
		differenced regression."""
		if hasattr(self,'Y_st'):
			return		
		m=panel.lost_obs
		N,T,k=panel.X.shape
		if model_parser.DEFAULT_INTERCEPT_NAME in panel.args.caption_d['beta']:
			m=self.args.args_d['beta'][0,0]
		else:
			m=panel.mean(panel.Y)	
		#e_norm=self.standardize_variable(panel,self.u,reverse_difference)
		self.Y_st,   self.Y_st_long   = self.standardize_variable(panel,panel.Y,reverse_difference)
		self.X_st,   self.X_st_long   = self.standardize_variable(panel,panel.X,reverse_difference)
		self.XIV_st, self.XIV_st_long = self.standardize_variable(panel,panel.XIV,reverse_difference)
		self.Y_pred_st=cf.dot(self.X_st,self.args.args_d['beta'])
		self.Y_pred=cf.dot(panel.X,self.args.args_d['beta'])	
		self.e_norm_long=self.stretch_variable(panel,self.e_norm)
		self.Y_pred_st_long=self.stretch_variable(panel,self.Y_pred_st)
		self.Y_pred_long=np.dot(panel.input.X,self.args.args_d['beta'])
		self.e_long=panel.input.Y-self.Y_pred_long
		
		Rsq, Rsqadj, LL_ratio,LL_ratio_OLS=stat_functions.goodness_of_fit(self, False, panel)
		Rsq2, Rsqadj2, LL_ratio2,LL_ratio_OLS2=stat_functions.goodness_of_fit(self, True, panel)
		a=0
				
	
	def standardize_variable(self,panel,X,norm=False,reverse_difference=False):
		X=panel.arma_dot.dot(self.AMA_1AR,X,self)
		X=(X+self.re_obj_i.RE(X, panel,False)+self.re_obj_t.RE(X, panel,False))
		if (not panel.Ld_inv is None) and reverse_difference:
			X=cf.dot(panel.Ld_inv,X)*panel.a[3]		
		if norm:
			X=X*self.v_inv05
		X_long=self.stretch_variable(panel,X)
		return X,X_long		
	
	def stretch_variable(self,panel,X):
		N,T,k=X.shape
		m=panel.map
		NT=panel.total_obs
		X_long=np.zeros((NT,k))
		X_long[m]=X
		return X_long
		
		

	def copy_args_d(self):
		return copy_array_dict(self.args.args_d)

	
	def h(self,panel,e,z):
		return h(e, z, panel)
	
	def arma_calc(self,panel, u):
		matrices=set_garch_arch(panel,self.args.args_d, u)
		if matrices is None:
			return None		
		self.AMA_1,self.AMA_1AR,self.GAR_1,self.GAR_1MA, self.e, self.var, h=matrices
		self.AMA_dict={'AMA_1':None,'AMA_1AR':None,'GAR_1':None,'GAR_1MA':None}		
		return matrices	
	


def set_garch_arch(panel,args,u):
	"""Solves X*a=b for a where X is a banded matrix with 1 or zero, and args along
	the diagonal band"""
	N, T, _ = u.shape
	rho=np.insert(-args['rho'],0,1)
	psi=args['psi']
	psi=np.insert(args['psi'],0,0) 

	
	AMA_1,AMA_1AR,GAR_1,GAR_1MA, e, var, h=(
		np.append([1],np.zeros(T-1)),
		np.zeros(T),
		np.append([1],np.zeros(T-1)),
		np.zeros(T),
		np.zeros(u.shape),
		np.zeros(u.shape),
		np.zeros(u.shape)
	)
	

	#c.arma_arrays(args['lambda'],rho,-args['gamma'],psi,T,AMA_1,AMA_1AR,GAR_1,GAR_1MA,u,e,var)	

	lmbda = args['lambda']
	gmma = -args['gamma']
	parameters = np.array(( N , T , 
							len(lmbda), len(rho), len(gmma), len(psi), 
							panel.options.EGARCH.value, panel.tot_lost_obs), 
							dtype=int)

	cfunct.armas(parameters.ctypes.data_as(CIPT), 
						  lmbda.ctypes.data_as(CDPT), rho.ctypes.data_as(CDPT),
						  gmma.ctypes.data_as(CDPT), psi.ctypes.data_as(CDPT),
						  AMA_1.ctypes.data_as(CDPT), AMA_1AR.ctypes.data_as(CDPT),
						  GAR_1.ctypes.data_as(CDPT), GAR_1MA.ctypes.data_as(CDPT),
						  u.ctypes.data_as(CDPT), 
						  e.ctypes.data_as(CDPT), 
						  var.ctypes.data_as(CDPT),
						  h.ctypes.data_as(CDPT)
						  )		
	

	r=[]
	#Creating nympy arrays with name properties. 
	for i in ['AMA_1','AMA_1AR','GAR_1','GAR_1MA']:
		r.append((locals()[i],i))
	for i in ['e', 'var', 'h']:
		r.append(locals()[i])
	return r


def set_garch_arch_scipy(panel,args):
	#after implementing ctypes, the scipy version might be dropped entirely
	p,q,d,k,m=panel.pqdkm
	nW,n=panel.nW,panel.max_T

	AAR=-lag_matr(-panel.I,args['rho'])
	AMA_1AR,AMA_1=solve_mult(args['lambda'], AAR, panel.I)
	if AMA_1AR is None:
		return
	GMA=lag_matr(panel.I*0,args['psi'])
	GAR_1MA,GAR_1=solve_mult(-args['gamma'], GMA, panel.I)
	if GAR_1MA is None:
		return
	r=[]
	for i in ['AMA_1','AMA_1AR','GAR_1','GAR_1MA']:
		r.append((locals()[i],i))
	return r
	
def solve_mult(args,b,I):
	"""Solves X*a=b for a where X is a banded matrix with 1  and args along
	the diagonal band"""
	import scipy
	n=len(b)
	q=len(args)
	X=np.zeros((q+1,n))
	X[0,:]=1
	X2=np.zeros((n,n))
	w=np.zeros(n)
	r=np.arange(n)	
	for i in range(q):
		X[i+1,:n-i-1]=args[i]
	try:
		X_1=scipy.linalg.solve_banded((q,0), X, I)
		if np.any(np.isnan(X_1)):
			return None,None			
		X_1b=cf.dot(X_1, b)
	except:
		return None,None

	return X_1b,X_1


def add_to_matrices(X_1,X_1b,a,ab,r):
	for i in range(0,len(a)):	
		if i>0:
			d=(r[i:],r[:-i])
			X_1[d]=a[i]
		else:
			d=(r,r)
		X_1b[d]=ab[i]	
	return X_1,X_1b

def lag_matr(L,args):
	k=len(args)
	if k==0:
		return L
	L=1*L
	r=np.arange(len(L))
	for i in range(k):
		d=(r[i+1:],r[:-i-1])
		if i==0:
			d=(r,r)
		L[d]=args[i]

	return L



def copy_array_dict(d):
	r=dict()
	for i in d:
		r[i]=np.array(d[i])
	return r


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import calculus_functions as cf
import calculus_ll as cll
import numpy as np
import time
import os

class gradient:
	
	def __init__(self,panel,callback):
		self.panel=panel
		self.callback=callback
		
	def arima_grad(self,k,x,ll,sign,pre):
		if k==0:
			return None
		(N,T,m)=x.shape
		x=self.panel.arma_dot.dotroll(pre,k,sign,x,ll)
		x.resize(N,T,k)
		extr_value=1e+100
		if np.max(np.abs(x))>extr_value:
			x[np.abs(x)>extr_value]=np.sign(x[np.abs(x)>extr_value])*extr_value
		return x*self.panel.a[3]

	def garch_arima_grad(self,ll,dRE,varname):
		panel=self.panel
		groupeffect=0
		groupeffect, dvRE_dx=None, None
		d_input=0
		if self.panel.N>1 and panel.options.fixed_random_group_eff.value>0 and not dRE is None:
			d_eRE_sq=2*ll.e_RE*dRE
			dmeane2=panel.mean(d_eRE_sq,(0,1))
			d_input=(d_eRE_sq-dmeane2)*panel.a[3]
			dvRE_dx=dmeane2*panel.a[3]-ll.re_obj_i_v.dRE(d_input,ll.varRE_input,varname,panel)-ll.re_obj_t_v.dRE(d_input,ll.varRE_input,varname,panel)
			groupeffect=ll.dvarRE*dvRE_dx*panel.a[3]
		

		dvar_sigma_G=None
		if self.panel.pqdkm[4]>0 and not dRE is None: 			#eqs. 33-34
			((N,T,k))=dRE.shape
			x=cf.prod((ll.h_e_val,dRE))	
			dvar_sigma_G=self.panel.arma_dot.dot(ll.GAR_1MA,x,ll)
		dvar_e=cf.add((dvar_sigma_G,groupeffect),True)
		return dvar_e,dvar_sigma_G,dvRE_dx,d_input


	def get(self,ll,DLL_e=None,dLL_var=None,return_G=False):
		

		self.callback(perc = 0.05, text = '', task = 'gradient')
		(self.DLL_e, self.dLL_var)=(DLL_e, dLL_var)
		panel=self.panel
		incl=self.panel.included[3]
		re_obj_i,re_obj_t=ll.re_obj_i,ll.re_obj_t
		u, e_RE,u_RE,h_e_val,var_ARMA,h_val,v=ll.u, ll.e_RE,ll.u_RE,ll.h_e_val,ll.var_ARMA,ll.h_val,ll.v
		p,q,d,k,m=panel.pqdkm
		nW=panel.nW
		if DLL_e is None:
			dLL_var, DLL_e=cll.gradient(ll,self.panel)
		self.X_RE = (panel.XIV+re_obj_i.RE(panel.XIV, panel)+re_obj_t.RE(panel.XIV, panel))*panel.included[3]
		#ARIMA:
		de_rho_RE=self.arima_grad(p,u_RE,ll,-1,ll.AMA_1)
		de_lambda_RE=self.arima_grad(q,e_RE,ll,-1,ll.AMA_1)
		de_beta_RE=-self.panel.arma_dot.dot(ll.AMA_1AR,self.X_RE,ll)*panel.a[3]
		
		(self.de_rho_RE,self.de_lambda_RE,self.de_beta_RE)=(de_rho_RE,de_lambda_RE,de_beta_RE)		
		
		dvar_sigma_rho,		dvar_sigma_rho_G,		dvRE_rho	, d_rho_input		=	self.garch_arima_grad(ll,	self.de_rho_RE,		'rho')
		dvar_sigma_lambda, 	dvar_sigma_lambda_G,	dvRE_lambda	, d_lambda_input	=	self.garch_arima_grad(ll,	self.de_lambda_RE,	'lambda')
		dvar_sigma_beta,	dvar_sigma_beta_G,		dvRE_beta	, d_beta_input		=	self.garch_arima_grad(ll,	self.de_beta_RE,	'beta')

		
		(self.dvar_sigma_rho,self.dvar_sigma_lambda,self.dvar_sigma_beta)=(dvar_sigma_rho,dvar_sigma_lambda,dvar_sigma_beta)
		(self.dvar_sigma_rho_G,self.dvar_sigma_lambda_G,self.dvar_sigma_beta_G)=(dvar_sigma_rho_G,dvar_sigma_lambda_G,dvar_sigma_beta_G)
		(self.dvRE_rho,self.dvRE_lambda,self.dvRE_beta)=(dvRE_rho,dvRE_lambda,dvRE_beta)
		(self.d_rho_input,self.d_lambda_input,self.d_beta_input)=(d_rho_input,d_lambda_input,d_beta_input)

		#GARCH:

		(dvar_gamma, dvar_psi, dvar_mu, dvar_z_G, dvar_z)=(None,None,None,None,None)
		if panel.N>1:
			dvar_mu=cf.prod((ll.dvarRE_mu,incl))
		else:
			dvar_mu=None	
			
		if m>0:
			dvar_gamma=self.arima_grad(k,var_ARMA,ll,1,ll.GAR_1)
			dvar_psi=self.arima_grad(m,h_val,ll,1,ll.GAR_1)
			if not ll.h_z_val is None:
				dvar_z_G=cf.dot(ll.GAR_1MA,ll.h_z_val)
				(N,T,k)=dvar_z_G.shape

			dvar_z=dvar_z_G


		(self.dvar_gamma, self.dvar_psi,self.dvar_mu,self.dvar_z_G,self.dvar_z)=(dvar_gamma, dvar_psi, dvar_mu, dvar_z_G, dvar_z)

		#LL

		
		#final derivatives:
		dLL_beta=cf.add((cf.prod((dvar_sigma_beta,dLL_var)),cf.prod((self.de_beta_RE,DLL_e))),True)
		dLL_rho=cf.add((cf.prod((dvar_sigma_rho,dLL_var)),cf.prod((self.de_rho_RE,DLL_e))),True)
		dLL_lambda=cf.add((cf.prod((dvar_sigma_lambda,dLL_var)),cf.prod((self.de_lambda_RE,DLL_e))),True)
		dLL_gamma=cf.prod((dvar_gamma,dLL_var))
		dLL_psi=cf.prod((dvar_psi,dLL_var))
		self.dvar_omega=panel.W_a
		dLL_omega=cf.prod((self.dvar_omega,dLL_var))
		dLL_mu=cf.prod((self.dvar_mu,dLL_var))
		dLL_z=cf.prod((self.dvar_z,dLL_var))
		

		G=cf.concat_marray((dLL_beta,dLL_rho,dLL_lambda,dLL_gamma,dLL_psi,dLL_omega,dLL_mu,dLL_z))
		g=np.sum(G,(0,1))
		#For debugging:
		#print (g)
		#import debug
		#gn=debug.grad_debug(ll,panel,0.00001)#debugging
		#if np.sum((g-gn)**2)>10000000:
		#	a=0
		#print(gn)
		#a=debug.grad_debug_detail(ll, panel, 0.00000001, 'LL', 'beta',0)
		#dLLeREn,deREn=debug.LL_calc_custom(ll, panel, 0.0000001)

		self.callback(perc = 0.08, text = '', task = 'gradient')
		

		
		if return_G:
			return  g,G
		else:	
			return g

class hessian:
	def __init__(self,panel,g,callback):
		self.panel=panel
		self.its=0
		self.g=g
		self.callback=callback
		
	
	def get(self,ll,d2LL_de2,d2LL_dln_de,d2LL_dln2):	
		H = self.hessian(ll,d2LL_de2,d2LL_dln_de,d2LL_dln2)
		return H


	def hessian(self,ll,d2LL_de2,d2LL_dln_de,d2LL_dln2):
		panel=self.panel
		tic=time.perf_counter()
		g=self.g
		p,q,d,k,m=panel.pqdkm
		incl=self.panel.included[3]
		
		GARM=(ll.GAR_1,m,1)
		
		GARK=(ll.GAR_1,k,1)

		d2var_gamma2		=   cf.prod((2, 
		                        cf.dd_func_lags(panel,ll,GARK, 	g.dvar_gamma,						g.dLL_var,  transpose=True)))
		d2var_gamma_psi		=	cf.dd_func_lags(panel,ll,GARK, 	g.dvar_psi,							g.dLL_var)

		d2var_gamma_rho		=	cf.dd_func_lags(panel,ll,GARK,	g.dvar_sigma_rho_G,						g.dLL_var)
		d2var_gamma_lambda	=	cf.dd_func_lags(panel,ll,GARK, 	g.dvar_sigma_lambda_G,					g.dLL_var)
		d2var_gamma_beta	=	cf.dd_func_lags(panel,ll,GARK, 	g.dvar_sigma_beta_G,					g.dLL_var)
		d2var_gamma_z		=	cf.dd_func_lags(panel,ll,GARK, 	g.dvar_z_G,							g.dLL_var)
		self.callback(perc = 0.2, text = '', task = 'hessian')
		d2var_psi_rho		=	cf.dd_func_lags(panel,ll,GARM, 	cf.prod((ll.h_e_val,g.de_rho_RE)),		g.dLL_var)
		d2var_psi_lambda	=	cf.dd_func_lags(panel,ll,GARM, 	cf.prod((ll.h_e_val,g.de_lambda_RE)),	g.dLL_var)
		d2var_psi_beta		=	cf.dd_func_lags(panel,ll,GARM, 	cf.prod((ll.h_e_val,g.de_beta_RE)),	g.dLL_var)
		d2var_psi_z			=	cf.dd_func_lags(panel,ll,GARM, 	ll.h_z_val,								g.dLL_var)

		AMAq=(ll.AMA_1,q,-1)
		d2var_lambda2,		d2e_lambda2		=	cf.dd_func_lags_mult(panel,ll,g,AMAq,	'lambda_RE',	'lambda_RE', transpose=True)
		d2var_lambda_rho,	d2e_lambda_rho	=	cf.dd_func_lags_mult(panel,ll,g,AMAq,	'lambda_RE',	'rho_RE' )
		d2var_lambda_beta,	d2e_lambda_beta	=	cf.dd_func_lags_mult(panel,ll,g,AMAq,	'lambda_RE',	'beta_RE')

		AMAp=(ll.AMA_1,p,-1)
		d2var_rho_beta,		d2e_rho_beta	=	cf.dd_func_lags_mult(panel,ll,g,AMAp,	'rho_RE',		'beta_RE', u_gradient=True)
		
		self.callback(perc = 0.4, text = '', task = 'hessian')
		
		d2var_mu_rho,d2var_mu_lambda,d2var_mu_beta,d2var_mu_z,mu=None,None,None,None,None
		if panel.N>1:
			d2var_mu_rho			=	cf.sumNT(cf.prod((ll.ddvarRE_mu_vRE, 	g.dvRE_rho,  	 	g.dLL_var)))
			d2var_mu_lambda			=	cf.sumNT(cf.prod((ll.ddvarRE_mu_vRE, 	g.dvRE_lambda,  	g.dLL_var)))
			d2var_mu_beta			=	cf.sumNT(cf.prod((ll.ddvarRE_mu_vRE, 	g.dvRE_beta,  	 	g.dLL_var)))
			d2var_mu_z=None
			d2var_mu2=0

		self.callback(perc = 0.5, text = '', task = 'hessian')
		d2var_z2				=	cf.dd_func_lags(panel,ll,ll.GAR_1MA, ll.h_2z_val,						g.dLL_var) 
		d2var_z_rho				=	cf.dd_func_lags(panel,ll,ll.GAR_1MA, cf.prod((ll.h_ez_val,g.de_rho_RE)),	g.dLL_var) 
		d2var_z_lambda			=	cf.dd_func_lags(panel,ll,ll.GAR_1MA, cf.prod((ll.h_ez_val,g.de_lambda_RE)),g.dLL_var) 
		d2var_z_beta			=	cf.dd_func_lags(panel,ll,ll.GAR_1MA, cf.prod((ll.h_ez_val,g.de_beta_RE)),	g.dLL_var) 
		
		d2var_rho2,	d2e_rho2	=	cf.dd_func_lags_mult(panel,ll,g,	None,	'rho_RE',		'rho_RE' )
		d2var_beta2,d2e_beta2	=	cf.dd_func_lags_mult(panel,ll,g,	None,	'beta_RE',		'beta_RE')
		


		(de_rho_RE,de_lambda_RE,de_beta_RE)=(g.de_rho_RE,g.de_lambda_RE,g.de_beta_RE)
		(dvar_sigma_rho,dvar_sigma_lambda,dvar_sigma_beta)=(g.dvar_sigma_rho,g.dvar_sigma_lambda,g.dvar_sigma_beta)
		(dvar_mu,dvar_z)=(g.dvar_mu, g.dvar_z)		

		d2var_beta_omega, d2var_rho_omega, d2var_lambda_omega=None, None, None
			
		self.callback(perc = 0.6, text = '', task = 'hessian')
		#Final:
		D2LL_beta2			=	cf.dd_func(d2LL_de2,	d2LL_dln_de,	d2LL_dln2,	de_beta_RE, 	de_beta_RE,		dvar_sigma_beta, 	dvar_sigma_beta,	d2e_beta2, 					d2var_beta2)
		D2LL_beta_rho		=	cf.dd_func(d2LL_de2,	d2LL_dln_de,	d2LL_dln2,	de_beta_RE, 	de_rho_RE,		dvar_sigma_beta, 	dvar_sigma_rho,		T(d2e_rho_beta), 		T(d2var_rho_beta))
		D2LL_beta_lambda	=	cf.dd_func(d2LL_de2,	d2LL_dln_de,	d2LL_dln2,	de_beta_RE, 	de_lambda_RE,	dvar_sigma_beta, 	dvar_sigma_lambda,	T(d2e_lambda_beta), 	T(d2var_lambda_beta))
		D2LL_beta_gamma		=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_beta_RE, 	None,			dvar_sigma_beta, 	g.dvar_gamma,		None, 					T(d2var_gamma_beta))
		D2LL_beta_psi		=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_beta_RE, 	None,			dvar_sigma_beta, 	g.dvar_psi,			None, 					T(d2var_psi_beta))
		D2LL_beta_omega		=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_beta_RE, 	None,			dvar_sigma_beta, 	g.dvar_omega,		None, 					d2var_beta_omega)
		D2LL_beta_mu		=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_beta_RE, 	None,			dvar_sigma_beta, 	dvar_mu,			None, 					d2var_mu_beta)
		D2LL_beta_z			=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_beta_RE, 	None,			dvar_sigma_beta, 	dvar_z,				None, 					T(d2var_z_beta))
		
		D2LL_rho2			=	cf.dd_func(d2LL_de2,	d2LL_dln_de,	d2LL_dln2,	de_rho_RE, 		de_rho_RE,		dvar_sigma_rho, 	dvar_sigma_rho,		d2e_rho2, 					d2var_rho2)
		D2LL_rho_lambda		=	cf.dd_func(d2LL_de2,	d2LL_dln_de,	d2LL_dln2,	de_rho_RE, 		de_lambda_RE,	dvar_sigma_rho, 	dvar_sigma_lambda,	T(d2e_lambda_rho), 		T(d2var_lambda_rho))
		D2LL_rho_gamma		=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_rho_RE, 		None,			dvar_sigma_rho, 	g.dvar_gamma,		None, 					T(d2var_gamma_rho))	
		D2LL_rho_psi		=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_rho_RE, 		None,			dvar_sigma_rho, 	g.dvar_psi,			None, 					T(d2var_psi_rho))
		D2LL_rho_omega		=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_rho_RE, 		None,			dvar_sigma_rho, 	g.dvar_omega,		None, 					d2var_rho_omega)
		D2LL_rho_mu			=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_rho_RE, 		None,			dvar_sigma_rho, 	dvar_mu,			None, 					T(d2var_mu_rho))
		D2LL_rho_z			=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_rho_RE, 		None,			dvar_sigma_rho, 	dvar_z,				None, 					T(d2var_z_rho))
		
		D2LL_lambda2		=	cf.dd_func(d2LL_de2,	d2LL_dln_de,	d2LL_dln2,	de_lambda_RE, 	de_lambda_RE,	dvar_sigma_lambda, 	dvar_sigma_lambda,	T(d2e_lambda2), 		T(d2var_lambda2))
		D2LL_lambda_gamma	=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_lambda_RE, 	None,			dvar_sigma_lambda, 	g.dvar_gamma,		None, 					T(d2var_gamma_lambda))
		D2LL_lambda_psi		=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_lambda_RE, 	None,			dvar_sigma_lambda, 	g.dvar_psi,			None, 					T(d2var_psi_lambda))
		D2LL_lambda_omega	=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_lambda_RE, 	None,			dvar_sigma_lambda, 	g.dvar_omega,		None, 					d2var_lambda_omega)
		D2LL_lambda_mu		=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_lambda_RE, 	None,			dvar_sigma_lambda, 	dvar_mu,			None, 					T(d2var_mu_lambda))
		D2LL_lambda_z		=	cf.dd_func(None,		d2LL_dln_de,	d2LL_dln2,	de_lambda_RE, 	None,			dvar_sigma_lambda, 	dvar_z,				None, 					T(d2var_z_lambda))
		
		D2LL_gamma2			=	cf.dd_func(None,		None,			d2LL_dln2,	None, 			None,			g.dvar_gamma, 		g.dvar_gamma,		None, 					T(d2var_gamma2))
		D2LL_gamma_psi		=	cf.dd_func(None,		None,			d2LL_dln2,	None,			None,			g.dvar_gamma, 		g.dvar_psi,			None, 					d2var_gamma_psi)
		D2LL_gamma_omega	=	cf.dd_func(None,		None,			d2LL_dln2,	None, 			None,			g.dvar_gamma, 		g.dvar_omega,		None, 					None)
		D2LL_gamma_mu		=	cf.dd_func(None,		None,			d2LL_dln2,	None, 			None,			g.dvar_gamma, 		dvar_mu,			None, 					None)
		D2LL_gamma_z		=	cf.dd_func(None,		None,			d2LL_dln2,	None, 			None,			g.dvar_gamma, 		dvar_z,				None, 					d2var_gamma_z)
		
		D2LL_psi2			=	cf.dd_func(None,		None,			d2LL_dln2,	None,			None,			g.dvar_psi, 		g.dvar_psi,			None, 					None)
		D2LL_psi_omega		=	cf.dd_func(None,		None,			d2LL_dln2,	None, 			None,			g.dvar_psi, 		g.dvar_omega,		None, 					None)
		D2LL_psi_mu			=	cf.dd_func(None,		None,			d2LL_dln2,	None, 			None,			g.dvar_psi, 		dvar_mu,			None, 					None)
		D2LL_psi_z			=	cf.dd_func(None,		None,			d2LL_dln2,	None, 			None,			g.dvar_psi, 		dvar_z,				None, 					d2var_psi_z)
		
		D2LL_omega2			=	cf.dd_func(None,		None,			d2LL_dln2,	None, 			None,			g.dvar_omega, 		g.dvar_omega,		None, 					None)
		D2LL_omega_mu		=	cf.dd_func(None,		None,			d2LL_dln2,	None, 			None,			g.dvar_omega, 		g.dvar_mu,			None, 					None)
		D2LL_omega_z		=	cf.dd_func(None,		None,			d2LL_dln2,	None, 			None,			g.dvar_omega, 		g.dvar_z,			None, 					None)
		
		D2LL_mu2			=	cf.dd_func(None,		None,			d2LL_dln2,	None, 			None,			dvar_mu, 			dvar_mu,			None, 					None)
		D2LL_mu_z			=	cf.dd_func(None,		None,			d2LL_dln2,	None, 			None,			dvar_mu, 			dvar_z,				None, 					d2var_mu_z)
		
		D2LL_z2				=	cf.dd_func(None,		None,			d2LL_dln2,	None, 			None,			dvar_z, 			dvar_z,				None, 					d2var_z2)
		

		
		
		H= [[D2LL_beta2,			D2LL_beta_rho,		D2LL_beta_lambda,		D2LL_beta_gamma,	D2LL_beta_psi,		D2LL_beta_omega,	D2LL_beta_mu,	D2LL_beta_z		],
	        [T(D2LL_beta_rho),		D2LL_rho2,			D2LL_rho_lambda,		D2LL_rho_gamma,		D2LL_rho_psi,		D2LL_rho_omega,		D2LL_rho_mu,	D2LL_rho_z			],
	        [T(D2LL_beta_lambda),	T(D2LL_rho_lambda),	D2LL_lambda2,			D2LL_lambda_gamma,	D2LL_lambda_psi,	D2LL_lambda_omega,	D2LL_lambda_mu,	D2LL_lambda_z		],
	        [T(D2LL_beta_gamma),	T(D2LL_rho_gamma),	T(D2LL_lambda_gamma),	D2LL_gamma2,		D2LL_gamma_psi,		D2LL_gamma_omega, 	D2LL_gamma_mu,	D2LL_gamma_z		],
	        [T(D2LL_beta_psi),		T(D2LL_rho_psi),	T(D2LL_lambda_psi),		T(D2LL_gamma_psi),	D2LL_psi2,			D2LL_psi_omega, 	D2LL_psi_mu,	D2LL_psi_z			],
	        [T(D2LL_beta_omega),	T(D2LL_rho_omega),	T(D2LL_lambda_omega),	T(D2LL_gamma_omega),T(D2LL_psi_omega),	D2LL_omega2, 		D2LL_omega_mu,	D2LL_omega_z		], 
	        [T(D2LL_beta_mu),		T(D2LL_rho_mu),		T(D2LL_lambda_mu),		T(D2LL_gamma_mu),	T(D2LL_psi_mu),		T(D2LL_omega_mu), 	D2LL_mu2,		D2LL_mu_z			],
	        [T(D2LL_beta_z),		T(D2LL_rho_z),		T(D2LL_lambda_z),		T(D2LL_gamma_z),	T(D2LL_psi_z),		T(D2LL_omega_z), 	D2LL_mu_z,		D2LL_z2				]]
		self.callback(perc = 0.8, text = '', task = 'hessian')
		H=cf.concat_matrix(H)
		#for debugging:
		#import debug
		#Hn=debug.hess_debug(ll,panel,g,0.00000001)#debugging
		#v=debug.hess_debug_detail(ll,panel,0.0000001,'grp','beta','beta',0,0)
		#print (time.perf_counter()-tic)
		self.its+=1
		if np.any(np.isnan(H)):
			return None
		#print(H[0]/1e+11)
		return H
	



	
def T(x):
	if x is None:
		return None
	return x.T


#!/usr/bin/env python
# -*- coding: utf-8 -*-

#This module contains the argument class for the panel object

import stat_functions as stat
import numpy as np
import loglikelihood as logl
import random_effects as re
import calculus_functions as cf




class arguments:
	"""Sets initial arguments and stores static properties of the arguments"""
	def __init__(self,panel):
		p, q, d, k, m=panel.pqdkm
		self.initial_user_defined = False
		self.categories=['beta','rho','lambda','gamma','psi','omega']
		if panel.z_active:
			self.categories+=['z']
		self.mu_removed=True
		if not self.mu_removed:
			self.categories+=['mu']
		self.make_namevector(panel,p, q, k, m)
		initargs=self.initargs(p, d, q, m, k, panel)
		self.position_defs(initargs)
		self.set_init_args(panel,initargs)
		self.get_user_constraints(panel)
		
		
	def get_user_constraints(self,panel):
		e="User contraints must be a dict of dicts or a string evaluating to that, on the form of ll.args.dict_string. User constraints not applied"
		if type(panel.options.user_constraints.value)==dict:
			self.user_constraints=panel.options.user_constraints.value
		else:
			if panel.options.user_constraints.value is None or panel.options.user_constraints.value=='':
				self.user_constraints={}
				return
			try:
				self.user_constraints=eval(panel.options.user_constraints.value)
			except SyntaxError:
				print(f"Syntax error: {e}")
				self.user_constraints={}
				return			
			except:
				print(e)
				self.user_constraints={}
				return
		if not panel.z_active and 'z' in self.user_constraints:
			self.user_constraints.pop('z')	
			
		for grp in self.user_constraints:
			try:
				self.user_constraints[grp] = np.array(self.user_constraints[grp]).flatten()
				test = 1.0 * self.user_constraints[grp]
			except:
				print(f'Failed to apply user constraints for {grp}')
				self.user_constraints.pop('grp')
				
		for grp in self.user_constraints:
			if not grp in self.caption_d:
				print(f"Constraint on {grp} not applied, {grp} not in arguments")
			if len(self.caption_d[grp]) < len(self.user_constraints[grp]):
				print(f"Not all constraints on {grp} applied, more constraints than arguments for {grp}")
				
		

	def initargs(self,p,d,q,m,k,panel):

		args=dict()
		args['beta']=np.zeros((panel.X.shape[2],1))
		args['omega']=np.zeros((panel.W.shape[2],1))
		args['rho']=np.zeros(p)
		args['lambda']=np.zeros(q)
		args['psi']=np.zeros(m)
		args['gamma']=np.zeros(k)
		args['omega'][0][0]=0
		args['mu']=np.array([])
		args['z']=np.array([])			

		
		if m>0 and panel.z_active:
			args['z']=np.array([1e-09])	

		if panel.N>1 and not self.mu_removed:
			args['mu']=np.array([0.0001])			
			

		return args

	def set_init_args(self,panel,initargs=None, default = True):
		p, q, d, k, m=panel.pqdkm
		
		if initargs is None:
			initargs = self.initargs(p, d, q, m, k, panel)
			
		#de2=np.roll(e**2,1)-e**2
		#c=stat.correl(np.concatenate((np.roll(de2,1),de2),2),panel)[0,1]
		beta,omega = self.set_init_regression(initargs,panel, default)
		self.args_start=self.create_args(initargs,panel)
		self.args_init=self.create_args(initargs,panel)
		self.set_restricted_args(p, d, q, m, k,panel,omega,beta)
		self.n_args=len(self.args_init.args_v)
		

	def set_restricted_args(self,p, d, q, m, k, panel,omega,beta):
		args_restricted=self.initargs(p, d, q, m, k, panel)
		args_restricted['beta'][0][0]=np.mean(panel.Y)
		args_restricted['omega'][0][0]=np.log(panel.var(panel.Y))
		self.args_restricted=self.create_args(args_restricted,panel)
		
		args_OLS=self.initargs(p, d, q, m, k, panel)	
		args_OLS['beta']=beta
		args_OLS['omega'][0][0]=omega
		self.args_OLS=self.create_args(args_OLS,panel)
		
	
	def create_null_ll(self,panel):
		if not hasattr(self,'LL_OLS'):
			self.LL_OLS=logl.LL(self.args_OLS,panel).LL
			self.LL_null=logl.LL(self.args_restricted,panel).LL	
		
	def position_defs(self,initargs):
		"""Defines positions in vector argument"""

		self.positions=dict()
		self.positions_map=dict()#a dictionary of indicies containing the string name and sub-position of index within the category
		k=0
		for i in self.categories:
			n=len(initargs[i])
			rng=range(k,k+n)
			self.positions[i]=rng
			for j in rng:
				self.positions_map[j]=[0,i,j-k]#equation,category,relative position
			k+=n

	
	def conv_to_dict(self,args):
		"""Converts a vector argument args to a dictionary argument. If args is a dict, it is returned unchanged"""
		if type(args)==dict:
			return args
		if type(args)==list:
			args=np.array(args)			
		d=dict()
		k=0
		for i in self.categories:
			n=len(self.positions[i])
			rng=range(k,k+n)
			d[i]=np.array(args[rng])
			if i=='beta' or i=='omega':
				d[i]=d[i].reshape((n,1))
			k+=n
		return d


	def conv_to_vector(self,args):
		"""Converts a dict argument args to vector argument. if args is a vector, it is returned unchanged.\n
		If args=None, the vector of self.args_init is returned"""
		if type(args)==list or type(args)==np.ndarray:
			return np.array(args)
		v=np.array([])
		for i in self.categories:
			s=np.array(args[i])
			if len(s.shape)==2:
				s=s.flatten()
			if len(s)>0:
				v=np.concatenate((v,s))
		return v


	def make_namevector(self,panel,p, q, k, m):
		"""Creates a vector of the names of all regression varaibles, 
		including variables, ARIMA and GARCH terms. This defines the positions
		of the variables througout the estimation."""
		d, names_d = {}, {}
		captions=list(panel.input.X.keys())#copy variable names
		d['beta']=list(captions)
		c=[list(captions)]
		names = [f'x{i}' for i in range(panel.n_beta)]
		names_d['beta'] = list(names)
		add_names(p,'rho%s    AR    p','rho',d,c,captions, names, names_d)
		add_names(q,'lambda%s MA    q','lambda',d,c,captions, names, names_d)
		add_names(k,'gamma%s  GARCH k','gamma',d,c,captions, names, names_d)
		add_names(m,'psi%s    ARCH  m','psi',d,c,captions, names, names_d)
		
		
		d['omega']=list(panel.input.W.keys())
		captions.extend(panel.input.W.keys())
		
		names_d['omega'] = [f'omega{i}' for i in range(panel.nW)]
		names.extend(names_d['omega'])
		
		c.append(d['omega'])
		
		if m>0:
			if panel.N>1 and not self.mu_removed:
				d['mu']=['mu (var.ID eff.)']
				captions.extend(d['mu'])
				names_d['mu']=['mu']
				names.extend(d['mu'])				
				c.append(d['mu'])
			if panel.z_active:
				d['z']=['z in h(e,z)']
				captions.extend(d['z'])
				names_d['z']=['z']
				names.extend(d['z'])					
				c.append(d['z'])
			
		self.caption_v=captions
		self.caption_d=d
		self.names_v = names
		self.names_d = names_d
		self.names_category_list=c

	def create_args(self,args,panel,constraints=None):
		if isinstance(args,arguments_set):
			self.test_consistency(args)
			return args
		args_v=self.conv_to_vector(args)
		if not constraints is None:
			constraints.within(args_v,True)	
			constraints.set_fixed(args_v)
		args_d=self.conv_to_dict(args_v)
		dict_string=[]
		for c in self.categories:
			s=[]
			captions=self.caption_d[c]
			a=args_d[c].flatten()
			for i in range(len(captions)):
				s.append(f"'{captions[i]}':{a[i]}")
			dict_string.append(f"'{c}':\n"+"{"+",\n".join(s)+"}")
		dict_string="{"+",\n".join(dict_string)+"}"
		return arguments_set(args_d, args_v, dict_string, self,panel)
	
	def test_consistency(self,args):
		#for debugging only
		m=self.positions_map
		for i in m:
			dict_arg=args.args_d[m[i][1]]
			if len(dict_arg.shape)==2:
				dict_arg=dict_arg[m[i][2]]
			if dict_arg[0]!=args.args_v[i]:
				raise RuntimeError("argument inconsistency")
			
	def get_name_ix(self,x,single_item=False):
		#returns name, list of indicies
		if x is None:
			return None, None
		if x in self.caption_v:
			if single_item:
				indicies=self.caption_v.index(x)
			else:
				indicies=[self.caption_v.index(x)]	
			return x,indicies
		elif x in self.positions and not single_item:
			indicies=list(self.positions[x])
			return x,indicies
		elif x in self.names_v:
			if single_item:
				indicies=self.names_v.index(x)
			else:
				indicies=[self.names_v.index(x)]	
			return x,indicies			
		try:
			name=self.caption_v[x]
		except Exception as e:
			raise RuntimeError(f"{e}. The identifier of an argument must be an integer or a string macthing a name in 'self.caption_v' or a category in 'self.positions'")
		if single_item:
			return name,x
		else:
			return name,[x]
		
	def set_init_regression(self, initargs,panel, default):
		usrargs =  panel.options.arguments.value
		if not usrargs is None:
			if type(usrargs)==str:
				try:
					usrargs = eval(usrargs.replace(" array"," np.array"))
				except NameError as e:
					if str(e)=="name 'array' is not defined":
						usrargs = eval(usrargs.replace("array"," np.array"))
			args = self.create_args(usrargs,panel)
			for c in args.args_d:
				initargs[c] = args.args_d[c]
			self.initial_user_defined = True
			return initargs['beta'], initargs['omega'][0,0]
		p, q, d, k, m=panel.pqdkm
		beta,rho,lmbda,u=ARMA_regression(panel)
		v=panel.var(u)
		if panel.options.EGARCH.value==0:
			omega0=panel.var(u)
			omega1 = 0
		else:
			omega0=np.log(panel.var(u))
			
		if panel.nW>1:
			omega1 = panel.var(panel.W[:,:,1:2])
			omega1 = omega0/omega1
			a=0.00
			initargs['omega'][1,0]=omega1*a
			initargs['omega'][0,0]=omega0*(1-a)
		else:
			initargs['omega'][0,0]=omega0
		initargs['beta']=beta
		
		if default: #turns out calculating ARMA parameters serves no purpose, default is therefore allways true. These coefficients are easily found by linesearch for given regression coefficients. 
			if q > 0:
				initargs['lambda'][0] = lmbda
				
			if p > 0:
				initargs['rho'][0] = rho	
	
		
		if panel.options.fixed_random_variance_eff.value==0:
			if v<1e-20:
				print('Warning, your model may be over determined. Check that you do not have the dependent among the independents')	
		return beta,omega0

def set_GARCH(panel,initargs,u,m):
	matrices=logl.set_garch_arch(panel,initargs)
	if matrices is None:
		e=u
	else:
		AMA_1,AMA_1AR,GAR_1,GAR_1MA=matrices
		e = cf.dot(AMA_1AR,u)*panel.included[3]		
	h=h_func(e, panel,initargs)
	if m>0:
		initargs['gamma'][0]=0
		initargs['psi'][0]=0


def h_func(e,panel,initargs):
	z=None
	if len(initargs['z'])>0:
		z=initargs['z'][0]	
	h_val,h_e_val,h_2e_val,h_z,h_2z,h_e_z=logl.h(e,z,panel)
	return h_val*panel.included[3]
	

	
def ARMA_regression(panel):
	gfre=panel.options.fixed_random_group_eff.value
	tfre=panel.options.fixed_random_time_eff.value
	re_obj_i=re.re_obj(panel,True,panel.T_i,panel.T_i,gfre)
	re_obj_t=re.re_obj(panel,False,panel.date_count_mtrx,panel.date_count,tfre)
	X=(panel.X+re_obj_i.RE(panel.X, panel)+re_obj_t.RE(panel.X, panel))*panel.included[3]
	Y=(panel.Y+re_obj_i.RE(panel.Y, panel)+re_obj_t.RE(panel.Y, panel))*panel.included[3]
	beta,u=stat.OLS(panel,X,Y,return_e=True)
	rho,lmbda=ARMA_process_calc(u,panel)
	return beta,rho,lmbda,u

def ARMA_process_calc(e,panel):
	c=stat.correlogram(panel,e,2,center=True)[1:]
	
	rho = max(min(c[1]/c[0],0.99), -0.99)

	
	lmbda = 0
	try:
		lmbda1 = (1 - 2*c[0]*rho + rho**2)/(2*(c[0]-rho))
		lmbda2 = (( (rho**2 - 1)*(rho**2 - 1 + 4*c[0]**2 - 4*c[0]*rho) )**0.5) / (2*(c[0]-rho))
		
		if abs(lmbda1+lmbda2)>abs(lmbda1-lmbda2):
			lmbda = max(min(lmbda1 - lmbda2,0.99), -0.99)
		else:
			lmbda = max(min(lmbda1 + lmbda2,0.99), -0.99)
	except:
		pass

	return rho,lmbda
		
			
def add_names(T,captionstr,category,d,c,captions, names, names_d):
	a=[]
	n=[]
	if ' ' in captionstr:
		namestr = captionstr.split(' ')[0]
	for i in range(T):
		a.append(captionstr %(i,))
		n.append(namestr %(i,))
	captions.extend(a)
	names.extend(n)
	d[category]=a
	names_d[category]=n
	c.append(a)


class arguments_set:
	"""A class that contains the numeric arguments used in the maximization
	in all shapes and forms needed."""
	def __init__(self,args_d,args_v,dict_string,arguments,panel):
		self.args_d=args_d#dictionary of arguments
		self.args_v=args_v#vector of arguments
		self.dict_string=dict_string#a string defining a dictionary of named arguments. For user input of initial arguments
		self.caption_v=arguments.caption_v#vector of captions
		self.caption_d=arguments.caption_d#dict of captions
		self.names_v=arguments.names_v#vector of names
		self.names_d=arguments.names_d#dict of names		
		self.n_args=len(self.args_v)
		self.pqdkm=panel.pqdkm
		self.positions=arguments.positions
		self.names_category_list=arguments.names_category_list
		


#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
MIN_DEN = 0

def LL(panel,var,e_REsq, e_RE):
	incl=panel.included[3]
	
	LL_const=-0.5*np.log(2*np.pi)
	if panel.options.EGARCH.value==0:
		MIN_LNV = 1e-30
		MAX_LNV = 1e+30
		
		a,k=panel.options.GARCH_assist.value, panel.options.kurtosis_adj.value
		
		dvar_pos=(var<MAX_LNV)*(var>MIN_LNV)
		var = incl*np.maximum(np.minimum(var,MAX_LNV),MIN_LNV)
		v=var	
		v_inv = incl/(var + MIN_DEN + (incl==0))	

		LL_full = LL_const-0.5*(incl*np.log(var+MIN_DEN + (incl==0))+(1-k)*e_REsq*v_inv
								+ a* (np.abs(e_REsq-var)*v_inv)
								+ (k/3)* e_REsq**2*v_inv**2
								)
	else:
		MIN_LNV = -100
		MAX_LNV = 100
		
		dvar_pos=(var < MAX_LNV) * (var > MIN_LNV)
		var = np.maximum(np.minimum(var, MAX_LNV), MIN_LNV)
		v = np.exp(var)*incl
		v_inv = np.exp(-var)*incl		
		LL_full = LL_const-0.5*(var+(e_REsq)*v_inv)
	return LL_full,v,v_inv,dvar_pos


def gradient(ll,panel):
	incl=panel.included[3]
	a,k=panel.options.GARCH_assist.value, panel.options.kurtosis_adj.value
	var,e_REsq,e_RE,v_inv=ll.var, ll.e_REsq, ll.e_RE,ll.v_inv 
	
	if panel.options.EGARCH.value==0:
		DLL_e   =-0.5*(	(1-k)*2*e_RE*v_inv	)
		dLL_var =-0.5*(	v_inv-(1-k)*(e_REsq)*v_inv**2	)

		DLL_e +=-0.5*(		
			  a* 2*np.sign(e_REsq-var)*e_RE*v_inv
			+ (k/3)* 4*e_REsq*e_RE*v_inv**2
				 )
		dLL_var +=-0.5*(	
				- a* (np.sign(e_REsq-var)*v_inv)
						- a* (np.abs(e_REsq-var)*v_inv**2)
				- (k/3)* 2*e_REsq**2*v_inv**3
				 )
	else:	
		DLL_e=-(ll.e_RE*ll.v_inv)
		dLL_var=-0.5*(incl-(ll.e_REsq*ll.v_inv)*incl)	
	dLL_var*=ll.dvar_pos*incl	
	DLL_e*=incl
	
	return dLL_var, DLL_e

def hessian(ll,panel):
	incl=panel.included[3]
	var,e_REsq,e_RE,v_inv=ll.var, ll.e_REsq, ll.e_RE,ll.v_inv 
	a,k=panel.options.GARCH_assist.value, panel.options.kurtosis_adj.value
	
	if panel.options.EGARCH.value==0:	
		d2LL_de2 	=-0.5*(	(1-k)*2*v_inv	)
		d2LL_dln_de =-0.5*(	-(1-k)*2*e_RE*v_inv**2)
		d2LL_dln2 	=-0.5*(-1*v_inv**2+(1-k)*2*(e_REsq)*v_inv**3	)
		
		d2LL_de2 	+=-0.5*(		
				  a* 2*np.sign(e_REsq-var)*v_inv
				+ (k/3)* 12*e_REsq*v_inv**2
					 )
		d2LL_dln_de +=-0.5*(		
				- a* 2*np.sign(e_REsq-var)*e_RE*v_inv**2
				- (k/3)* 8*e_REsq*e_RE*v_inv**3
					 )
		d2LL_dln2 	+=-0.5*(	
					+ a* (np.sign(e_REsq-var)*v_inv**2)
							+ a* 2*(np.abs(e_REsq-var)*v_inv**3)
							+ a* (np.sign(e_REsq-var)*v_inv**2)
					+ (k/3)* 6*e_REsq**2*v_inv**4
					 )

	else:
		d2LL_de2=-ll.v_inv
		d2LL_dln_de=ll.e_RE*ll.v_inv
		d2LL_dln2=-0.5*ll.e_REsq*ll.v_inv
	d2LL_dln_de*=ll.dvar_pos*incl
	d2LL_dln2*=ll.dvar_pos*incl
	d2LL_de2*=incl
	
	return d2LL_de2, d2LL_dln_de, d2LL_dln2




#!/usr/bin/env python
# -*- coding: utf-8 -*-

#This module handle interfacing for various output paltforms


try:
	get_ipython()#fails if not in IPython environment
	import IPython
except:
	IPython = None

import output
import os
import numpy as np
import time



WEB_PAGE='paneltime.html'
TMP_PAGE='tmphtml'
pic_num=[1]





def get_channel(window,exe_tab,panel, console_output):
	if console_output:
		return console(panel)
	if not window is None:#tkinter gui
		return tk_widget(window,exe_tab,panel)
	if not IPython is None:
		n=IPython.get_ipython().__class__.__name__
		if n=='ZMQInteractiveShell':
			return web_output(True,panel)
	try:
		return web_output(False,panel)
	except:
		pass
	return console(panel)
		
class web_output:
	def __init__(self,Jupyter,panel):
		global webbrowser
		import webbrowser
		self.panel=panel	
		self.Jupyter=Jupyter
		if not Jupyter:
			self.save_html(get_web_page('None', 'None', 'None', '', True))
			if panel.options.web_open_tab.value:
				webbrowser.open(WEB_PAGE, new = 2)
		self.output_set = False
		#import charts
		#self.charts = charts.ProcessCharts(panel)
			
	def set_progress(self,perc, text, task):
		return True
		
	def set_output_obj(self,ll, comput, dx_norm):
		"sets the outputobject in the output" 
		if self.output_set:
			return		
		self.output=output.output(ll,self.panel, comput, dx_norm)
		self.output_set = True
		
		
	def update(self,comput, its, ll, incr, dx_norm):
		self.output.update(comput,its, ll, incr, dx_norm)
		self.its=its
		self.reg_table=self.output.reg_table()
		tbl,llength=self.reg_table.table(4,'(','HTML',True,
							   show_direction=True,
							   show_constraints=True)		
		web_page=get_web_page(ll.LL, 
							  ll.args.args_v, 
							  dx_norm,
							  tbl,
							  self.Jupyter==False)
		#self.charts.save_all(ll)
		if self.Jupyter:
			IPython.display.clear_output(wait=True)
			display(IPython.display.HTML(web_page))
		else:
			self.save_html(web_page)

		
	def save_html(self,htm_str):
		self.f = open(WEB_PAGE, "w")
		self.f.write(htm_str)
		self.f.close()

		
	def print_final(self, msg, fret, conv, t0, xsol, its, node):
		print(msg)
		print(f"LL={fret}  success={conv}  t={time.time()-t0}  its: {its}   node: {node}")
		print(xsol)	
		
class console:
	def __init__(self,panel):
		self.panel=panel
		self.output_set = False
		
	def set_progress(self,perc,text, task):
		if task=='done':
			print(text)
		#perc = f'{int(perc*100)}%'.ljust(5)
		#print(f"{perc} - {task}: {text}")
		return True
		
	def set_output_obj(self,ll, comput, dx_norm):
		if self.output_set:
			return
		self.output=output.output(ll,self.panel, comput, dx_norm)
		self.output_set = True
		self.comput = comput

	def update(self,comput, its,ll,incr, dx_norm):
		self.ll = ll
		self.incr = incr
		self.dx_norm = dx_norm

		
	def print_final(self, msg, fret, conv, t0, xsol, its, node):
		if self.output_set:
			self.output.update(self.comput,its, self.ll, 0, self.dx_norm)
			self.its=its
			self.reg_table=self.output.reg_table()	
			tbl,llength=self.reg_table.table(5,'(','CONSOLE',False,
								   show_direction=False,
								   show_constraints=False)	
			print(tbl)
			return
		print(msg)
		print(f"LL={fret}  success={conv}  t={time.time()-t0}  its: {its}   node: {node}")
		print(xsol)		
				
class tk_widget:
	def __init__(self,window,exe_tab,panel):
		self.panel=panel
		self.tab=window.main_tabs._tabs.add_output(exe_tab)
		self.set_progress=self.tab.progress_bar.set_progress
		self.output_set = False
		
	def set_output_obj(self,ll, comput, dx_norm):
		if self.output_set:
			return		
		self.tab.set_output_obj(ll,self.panel, comput, dx_norm)
		self.output = self.tab.output
		self.output_set = True
		
	def update(self,comput, its, ll, incr, dx_norm):
		self.tab.update(self.panel, comput,its, ll, incr, dx_norm)
		
	def print_final(self, msg, fret, conv, t0, xsol, its, node):
		print(msg)
		print(f"LL={fret}  success={conv}  t={time.time()-t0} its: {its}")
		print(xsol)	

def get_web_page(LL, args, comput,tbl,auto_update):
	au_str=''
	if auto_update:
		au_str="""<meta http-equiv="refresh" content="1" >"""
	img_str=''
	pic_num[0]+=1
	if os.path.isfile('img/chart0.png'):
		img_str=(f"""<img src="img/histogram.png"><br>\n"""
				f"""<img src="img/correlogram.png"   ><br>\n"""
				f"""<img src="img/correlogram_variance.png"   >""")
	return f"""
<meta charset="UTF-8">
{au_str}
<head>
<title>paneltime output</title>
</head>
<style>
p {{
  margin-left: 60px;
  max-width: 980px;
  font-family: "Serif";
  text-align: left;
  color:#063f5c;
  font-size: 16;
}}
h1 {{
  margin-left: 20px;
  max-width: 980px;
  font-family: "Serif";
  text-align: left;
  color:black;
  font-size: 25;
  font-weight: bold;
}}

table.head {{
  font-family: "Serif";
  text-align: right;
  color:black;
  padding-left: 0px;
  font-size: 16;
}}
td.h:nth-child(odd) {{
  background: #CCC
}}
td {{
  padding-left: 0px;
}}
table {{
  font-family: "Serif";
  text-align: right;
  color:black;
  font-size: 16;
}}

th {{
border-collapse: collapse;
	border-bottom: double 3px;
	padding-left: 0px;
	white-space: nowrap;
}}
</style>
<body>
<div style='position:absolute;float:right;top:0;right:0'>
{img_str}
</div>
{tbl}
</body>
</html> """	




#!/usr/bin/env python
# -*- coding: utf-8 -*-
import calculus
import calculus_ll as cll
import numpy as np
import constraints
import loglikelihood as logl
import time
import stat_dist
import stat_functions as stat

EPS=3.0e-16 
TOLX=(4*EPS) 
STPMX=100.0 



class Computation:
	def __init__(self,panel, gtol, tolx, callback, nummerical, diag_hess):
		"""callback is a function taking any named arguments """
		if callback is None:
			callback = lambda **kwargs: None#function that does nothing
		self.callback = callback
		self.gradient=calculus.gradient(panel, self.callback)
		self.gtol = gtol
		self.tolx = tolx
		self.hessian=calculus.hessian(panel,self.gradient, self.callback)
		self.panel=panel
		self.constr=None
		self.CI=0
		self.weak_mc_dict={}
		self.mc_problems=[]
		self.H_correl_problem=False
		self.singularity_problems=False
		self.num_hess_count = 0
		self.H, self.g, self.G = None, None, None
		self.mcollcheck = False
		self.rec =[]
		self.nummerical = nummerical
		self.quit = False
		self.avg_incr = 0
		self.ev_constr = False
		self.diag_hess = diag_hess
		self.CI_anal = 0
		p, q, d, k, m = panel.pqdkm
		self.init_arma = 2#((p>1)|(q>1)|(k>1)|(m>1))*2
		if panel.args.initial_user_defined:
			self.init_arma = 0
		self.init_arma_its = 0



	def set(self, its,increment,lmbda,rev, add_dyn_constr, H, ll):
		self.its=its
		self.lmbda=lmbda
		self.has_reversed_directions=rev
		self.increment=increment
		self.constr_old=self.constr
		self.constr = constraints.Constraints(self.panel,ll.args,its)
			
		if self.constr is None:
			self.H_correl_problem,self.mc_problems,self.weak_mc_dict=False, [],{}
			self.singularity_problems=(len(self.mc_problems)>0) or self.H_correl_problem
			return
			
		if self.panel.options.constraints_engine.value:
			self.constr.add_static_constraints(self.panel, its, self.init_arma, ll)	
			self.constr.add_dynamic_constraints(self, H, ll)	
			
		self.CI=self.constr.CI
		self.H_correl_problem=self.constr.H_correl_problem	
		self.mc_problems=self.constr.mc_problems
		self.weak_mc_dict=self.constr.weak_mc_dict
		

	def exec(self, dx, dx_norm, hessin, H, f, x, g_old, incr, rev, alam, its, ll, calc = True):
		#Thhese setting may not hold for all circumstances, and should be tested properly:
		NUM_ITER = 5
		TOTP_TOL = 0.0001
		
		g, G = self.calc_gradient(ll)
		
		pgain, totpgain = potential_gain(dx, g, H)
		self.totpgain = totpgain
		max_pgain = abs(max(pgain))
		
		a = np.ones(len(g))
		if not self.constr is None:
			a[list(self.constr.fixed.keys())] =0		
		g_norm =np.max(np.abs(g*a*x)/(abs(f)+1e-12) )
		
		self.rec.append(str(i) for i in [totpgain, max_pgain, incr, g_norm])
		CI=0
		if not self.CI is None:
			CI = self.CI

		det = np.linalg.det(H)
		se = [None]*len(H)	
		if its >NUM_ITER and ((abs(g_norm) < self.gtol) or ((abs(totpgain)<TOTP_TOL) and its>5000)) and self.init_arma==0:
			Ha = self.calc_hessian(ll)
			self.constr.add_dynamic_constraints(self, Ha, ll)
			keep = [not i in self.constr.fixed.keys() for i in range(len(H))]
			Ha_keep = Ha[keep][:,keep]
			self.CI_anal = condition_index(Ha_keep)
			try:
				det = np.linalg.det(Ha_keep)
				hessin[keep][:,keep] = np.linalg.inv(Ha_keep)
				se = stat.robust_se(self.panel, 100, hessin, G)[0]				
			except Exception as e:
				print(e)
	
			if abs(g_norm) < self.gtol:
				return x, f, hessin, Ha, g, 1, se, det
			elif abs(totpgain)<TOTP_TOL:
				return x, f, hessin, Ha, g, 2, se, det
			
		#print(f"its:{its}, f:{f}, init_reg:{self.init_arma},gnorm: {abs(g_norm)}")

		if its>7 or abs(g_norm)<10*self.gtol:
			self.init_arma = 0
		elif (its>3 or abs(g_norm)<100*self.gtol) and not self.panel.args.initial_user_defined:
			self.init_arma = 1		

			
		self.avg_incr = incr + self.avg_incr*0.7
		self.ev_constr = False#self.CI>1000
		
		err = np.max(np.abs(dx)) < TOLX
		

			
		calchess = err or ((self.CI>10000) and (self.num_hess_count>10) and abs(g_norm)>100*self.gtol)#or (self.num_hess_count> and its <30) #or ((self.CI>10000) and (self.num_hess_count>1))
	
		if calc:
			if calchess:
				print('analytic')
				Ha = self.calc_hessian(ll)
				a = 0.5
				try:
					hessin = np.linalg.inv(Ha)*a + (1-a)*hessin
				except:
					pass
				self.num_hess_count = 0

			else:
				self.num_hess_count +=1
				try:
					hessin=self.hessin_num(hessin, g-g_old, dx)
					Hn = np.linalg.inv(hessin)
					H = Hn
				except:
					H = self.calc_hessian(ll)

		self.H, self.g, self.G = H, g, G
			
		self.set(its, incr, alam, rev, True, H, ll)
		
	
		
		return x, f, hessin, H, g, 0, se, det



	def calc_gradient(self,ll):
		dLL_lnv, DLL_e=cll.gradient(ll,self.panel)
		self.LL_gradient_tobit(ll, DLL_e, dLL_lnv)
		g, G = self.gradient.get(ll,DLL_e,dLL_lnv,return_G=True)
		return g, G
	

	def calc_hessian(self, ll):
		d2LL_de2, d2LL_dln_de, d2LL_dln2 = cll.hessian(ll,self.panel)
		self.LL_hessian_tobit(ll, d2LL_de2, d2LL_dln_de, d2LL_dln2)
		H = self.hessian.get(ll,d2LL_de2,d2LL_dln_de,d2LL_dln2)	
		return H
	
	def LL_gradient_tobit(self,ll,DLL_e,dLL_lnv):
		g=[1,-1]
		self.f=[None,None]
		self.f_F=[None,None]
		for i in [0,1]:
			if self.panel.tobit_active[i]:
				I=self.panel.tobit_I[i]
				self.f[i]=stat_dist.norm(g[i]*ll.e_norm[I], cdf = False)
				self.f_F[i]=(ll.F[i]!=0)*self.f[i]/(ll.F[i]+(ll.F[i]==0))
				self.v_inv05=ll.v_inv**0.5
				DLL_e[I]=g[i]*self.f_F[i]*self.v_inv05[I]
				dLL_lnv[I]=-0.5*DLL_e[I]*ll.e_RE[I]
				a=0
				

	def LL_hessian_tobit(self,ll,d2LL_de2,d2LL_dln_de,d2LL_dln2):
		g=[1,-1]
		if sum(self.panel.tobit_active)==0:
			return
		self.f=[None,None]
		e1s1=ll.e_norm
		e2s2=ll.e_REsq*ll.v_inv
		e3s3=e2s2*e1s1
		e1s2=e1s1*self.v_inv05
		e1s3=e1s1*ll.v_inv
		e2s3=e2s2*self.v_inv05
		f_F=self.f_F
		for i in [0,1]:
			if self.panel.tobit_active[i]:
				I=self.panel.tobit_I[i]
				f_F2=self.f_F[i]**2
				d2LL_de2[I]=      -(g[i]*f_F[i]*e1s3[I] + f_F2*ll.v_inv[I])
				d2LL_dln_de[I] =   0.5*(f_F2*e1s2[I]  +  g[i]*f_F[i]*(e2s3[I]-self.v_inv05[I]))
				d2LL_dln2[I] =     0.25*(f_F2*e2s2[I]  +  g[i]*f_F[i]*(e1s1[I]-e3s3[I]))
				
	
	def hessin_num(self, hessin, dg, xi):				#Compute difference of gradients,
		n=len(dg)
		#and difference times current matrix:
		hdg=(np.dot(hessin,dg.reshape(n,1))).flatten()
		fac=fae=sumdg=sumxi=0.0 							#Calculate dot products for the denominators. 
		fac = np.sum(dg*xi) 
		fae = np.sum(dg*hdg)
		sumdg = np.sum(dg*dg) 
		sumxi = np.sum(xi*xi) 
		if (fac < (EPS*sumdg*sumxi)**0.5):  					#Skip update if fac not sufficiently positive.
			fac=1.0/fac 
			fad=1.0/fae 
															#The vector that makes BFGS different from DFP:
			dg=fac*xi-fad*hdg   
			#The BFGS updating formula:
			hessin+=fac*xi.reshape(n,1)*xi.reshape(1,n)
			hessin-=fad*hdg.reshape(n,1)*hdg.reshape(1,n)
			hessin+=fae*dg.reshape(n,1)*dg.reshape(1,n)		
			
		return hessin

	
	def init_ll(self,args=None, ll=None):
		if args is None:
			args = ll.args.args_v
		self.constr=constraints.Constraints(self.panel, args)
		self.constr.add_static_constraints(self.panel, 0, self.init_arma)			
		try:
			args=args.args_v
		except:
			pass#args must be a vector
		for i in self.constr.fixed:
			args[i]=self.constr.fixed[i].value
		if ll is None:
			ll=logl.LL(args, self.panel, constraints=self.constr,print_err=True)
		if ll.LL is None:
			print("WARNING: Initial arguments failed, attempting default OLS-arguments ...")
			self.panel.args.set_init_args(self.panel,default=True)
			ll=logl.LL(self.panel.args.args_OLS,self.panel,constraints=self.constr,print_err=True)
			if ll.LL is None:
				raise RuntimeError("OLS-arguments failed too, you should check the data")
			else:
				print("default OLS-arguments worked")
		return ll
	
	def calc_init_dir(self, p0, full = False):
		"""Calculates the initial computation""" 
		ll = self.init_ll(p0)
		g, G = self.calc_gradient(ll)
		H = self.calc_hessian(ll)
		if self.diag_hess:
			d = np.diag(H)
			d = d - (np.abs(d)<1e-100)
			H = np.diag(d)
			hessin = np.diag(1/d)
		else:
			hessin = np.linalg.inv(H)
		return p0, ll, ll.LL , g, hessin, H
	
def det_managed(H):
	try:
		return np.linalg.det(H)
	except:
		return 1e+100
	
def inv_hess(hessian):
	try:
		h=-np.linalg.inv(hessian)
	except:
		return None	
	return h

def condition_index(H):
	n = len(H)
	d=np.maximum(np.abs(np.diag(H)).reshape((n,1)),1e-30)**0.5
	C = -H/(d*d.T)
	ev = np.abs(np.linalg.eigvals(C))**0.5
	if min(ev) == 0:
		return 1e-150
	return max(ev)/min(ev)	

		
def hess_inv(h, hessin):
	try:
		h_inv = np.linalg.inv(h)
	except Exception as e:
		print(e)
		return hessin
	return h_inv



def potential_gain(dx, g, H):
	"""Returns the potential gain of including each variables, given that all other variables are included and that the 
	quadratic model is correct. An alternative convercence criteria"""
	n=len(dx)
	rng=np.arange(len(dx))
	dxLL=dx*0
	dxLL_full=(sum(g*dx)+0.5*np.dot(dx.reshape((1,n)),np.dot(H,dx.reshape((n,1)))))[0,0]
	for i in range(len(dx)):
		dxi=dx*(rng!=i)
		dxLL[i]=dxLL_full-(sum(g*dxi)+0.5*np.dot(dxi.reshape((1,n)),np.dot(H,dxi.reshape((n,1)))))[0,0]
		
	return dxLL, dxLL_full






#!/usr/bin/env python
# -*- coding: utf-8 -*-

#This module handle callbacks



class CallBack:
	def __init__(self, inbox, outbox):
		self.inbox = inbox
		self.outbox = outbox
		self.outbox['quit'] = False

		
	def callback(self, **keywordargs):
		#print(f'writing to {id(self.outbox)}')
		for k in keywordargs:
			self.outbox[k] = keywordargs[k]
		return dict(self.inbox)
		

		
			

			




	
	
	


import numpy as np
import time
import direction
import linesearch
import callback
#import stat_functions as stat


#This module finds the array of arguments that minimizes some function. The derivative 
#of the function also needs to be supplied. 
#This is an adaption of the Broyden-Fletcher-Goldfarb-Shanno variant of Davidon-Fletcher-Powell algorithm by 
#Press, William H,Saul A Teukolsky,William T Vetterling and Brian P Flannery. 1992. Numerical Recipes in C'. 
#Cambridge: Cambridge University Press.



EPS=3.0e-16 
TOLX=(4*EPS) 
GTOL = 1e-5

def dfpmax(x, comput, callback, panel, slave_id):
	"""Given a starting point x[1..n] that is a vector of length n, the Broyden-Fletcher-Goldfarb-
	Shanno variant of Davidon-Fletcher-Powell minimization is performed on a function func, using
	its gradient as calculated by a routine dfunc. The convergence requirement on zeroing the
	gradient is input as gtol. Returned quantities are x[1..n] (the location of the minimum),
	iter (the number of iterations that were performed), and fret (the minimum value of the
	function). The routine lnsrch is called to perform approximate line minimizations.
	fargs are fixed arguments that ar not subject to optimization. ("Nummerical Recipes for C") """

	x, ll, f, g, hessin, H = comput.calc_init_dir(x)

	its, msg = 0, ''
	MAXITER = 10000
	
	cbhandler = CallBackHandler(callback, slave_id)
	fdict = {}
	for its in range(MAXITER):  	#Main loop over the iterations.


		dx, dx_norm, H_ = direction.get(g, x, H, comput.constr, f, hessin, comput.ev_constr, simple=False)
		ls = linesearch.LineSearch(x, comput, panel)
		ls.lnsrch(x, f, g, dx)	

		dx = ls.x - x
		incr = ls.f - f
		fdict[its] = ls.f
		
		
		x, f, hessin, H, g, conv, se, det = comput.exec(dx, dx_norm,  hessin, H, ls.f, ls.x, ls.g, incr, ls.rev, ls.alam, its, ls.ll)
		
		err = np.max(np.abs(dx)) < TOLX
		
		terminate = (conv>0) or err or its+1==MAXITER
		
		#print(f"sid:{slave_id}, f:{ls.f}, conv:{conv}, its:{its}")

		if conv==1:
			msg = "Convergence on zero gradient; local or global minimum identified"
		elif conv==2:
			msg = "Convergence on zero expected gain; local or global minimum identified given multicolinearity constraints"		
		elif err:
			msg = "Warning: Convergence on delta x; the gradient is incorrect or the tolerance is set too low"
		elif terminate:
			msg = "No convergence within %s iterations" %(MAXITER,)
			

		cbhandler.assign(ls, msg, dx_norm, f, x, H, comput, g, hessin, dx, 
						  incr, its, 'linesearch', terminate, 
						  conv, fdict, se, det)			

		if terminate or cbhandler.quit:	
			if terminate:
				cause = msg
			else:
				cause = 'forced'
			print(f"quit slave {slave_id}, time: {time.time()}, cause: {cause}, conv:{conv}")
			return cbhandler.callback.outbox, ls.ll
			
class CallBackHandler:
	def __init__(self, callback, slave_id):
		self.t = time.time()
		self.callback = callback
		self.id = slave_id
		self.quit = False
		self.inbox = {}
		
															
	def assign(self, ls, msg, dx_norm, f, x, H, comput, g, hessin, dx, incr, its, 
						  task, terminate, conv, fdict, se, det):


		self.check_for_quit_order()
		


		if msg == '':
			msg = ls.msg	
			
		if time.time()-self.t<1 and not (self.quit or terminate):
			return
		self.t = time.time()

		self.inbox = self.callback.callback(msg = msg, dx_norm = dx_norm, f = f, x = x, 
				 H = H, G=comput.G, g = g, hessin = hessin, dx = dx, 
				 incr = incr, rev = ls.rev, alam = ls.alam, 
				 its = its, constr = comput.constr, perc=min(its/100, 1), task = task, 
				 terminate= terminate or self.quit, conv = conv, slave_id = self.id, 
				 fdict = dict(fdict), CI = comput.CI, CI_anal = comput.CI_anal, tpgain = comput.totpgain, se = se, det = det)
		
	def check_for_quit_order(self):

		if not 'quit' in  self.callback.inbox:
			return
		if not self.callback.inbox['quit']:
			return
		self.quit = True
		
	
				





#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
import subprocess
import pickle
import datetime
from queue import Queue
from threading import Thread
import time
import tempfile
import random
import string
import parallel_slave
import hashlib
import psutil
import signal
import numpy as np
import __main__

MIN_TIME = 1

class Parallel():
	"""creates the slaves"""
	def __init__(self, max_nodes, fpath, run_parallel, callback_active, direct, layer = 0):
		"""module is a string with the name of the modulel where the
		functions you are going to run are """
		if max_nodes is None:
			self.cpu_count=os.cpu_count()
		else:
			self.cpu_count=max_nodes
		self.callback_active = callback_active
		n=self.cpu_count
		self.kill_warned = False
		self.is_parallel = run_parallel
		self.layer = layer
		self.direct = direct #flag to indicate to the host process whether the process should be run in the
		#directely in the simplest way, without any layers or multiprocessing (for debugging).
		self.kill_orphans()
		self.dict_file = create_temp_files(self.cpu_count, fpath)
		self.fpath = makepath(fpath)
		self.slaves=[Slave(run_parallel, n) for i in range(n)]
		self.final_results = {}
		self.n_tasks = {}
		self.t = time.time()
		self.pids = []
		self.master_pid = os.getpid()
		self.sum_running = {}
		for i in range(n):
			self.slaves[i].confirm(i, self.fpath) 
			pid=str(self.slaves[i].p_id)
			if int(i/5.0)==i/5.0:
				pid='\n'+pid
			self.pids.append(pid)
		pstr="""Multi core processing enabled using %s cores. \n
Master PID: %s \n
Slave PIDs: %s"""  %(n, self.master_pid,', '.join(self.pids))
		print (pstr)

	def send_dict(self, d):
		f = open(self.dict_file, 'wb')
		pickle.dump(d, f)  
		f.close()
		r = []		
		return self.async_send_receive('transfer dictionary', self.dict_file)

			
	def exec(self, tasks, name):
		#Handles at most as many tasks as cpus for the moment
		if tasks is None:
			tasks = [None]*self.cpu_count
		if type(tasks)==str:
			tasks = [tasks]*self.cpu_count
		self.n_tasks[name] = len(tasks)
		self.final_results[name] = [None]*self.n_tasks[name]
		self.sum_running[name] = len(tasks)
		return self.async_send_receive(name, tasks)
			
	def callback(self, name, outbox = {}, s_id = None, collect_finnished_procs = True):
		if not self.callback_active:
			return [{}]*self.cpu_count
		d=[]
		sumrun = -1
		while time.time() - self.t<MIN_TIME or sumrun<0:
			is_running = np.array(self.check_state(name))
			sumrun = sum(is_running)
			if sumrun<self.sum_running[name] :
				self.sum_running[name] = sumrun
				break #break if changed
			time.sleep(0.1)
		self.t = time.time()		
			

		if not s_id is None:
			self.slaves[s_id].send('callback',(name,outbox))
			return self.slaves[s_id].receive()	
		
		if not collect_finnished_procs:
			return self.async_send_receive('callback', (name, outbox))
		
		response = self.final_results[name]
		for i in range(self.n_tasks[name]):
			if (not is_running[i]) and (response[i] is None):
				r = self.collect(name, i)
				if not r is None and False:#set to true for debug
					print(f"collected sid:{i}, conv:{r['conv']}, f:{r['f']}, layer: {self.layer}")
				response[i] = r
		remaining_procs = np.arange(self.n_tasks[name])[np.array([i is None for i in response])]
		if len(remaining_procs)==0:
			return response
		response = list(response)
		r = self.async_send_receive('callback', (name, outbox), remaining_procs)
		for i in range(self.n_tasks[name]):
			if response[i] is None:
				response[i] = r[i]
		return response


	def collect(self, name, sid=None):
		if not sid is None:
			self.slaves[sid].send('collect', name)
			msg = self.slaves[sid].receive()	
			return 	msg

		return self.async_send_receive('collect', name)
		
	def async_send_receive(self,msg, obj, slave_list=[]):
		r = []
		q = Queue()

		if not type(obj) == list:
			obj = [obj]*self.cpu_count
		if len(slave_list)==0:
			slave_list = range(len(obj))		
		for i in slave_list:
			self.slaves[i].send(msg, obj[i])
			t = Thread(target = self.slaves[i].receive, args = (q, ), daemon= True)
			t.start()
		r = [{}]*len(obj)
		for i in slave_list:
			res, sid = q.get()
			r[sid] = res
		return r
		
	def count_alive(self):
		return sum(self.exec(None, 'check_state'))
	
	def check_state(self, name = None):
		return self.exec(name, 'check_state')


			
	def quit(self):
		print('quitting')
		for i in self.slaves:
			i.p.stdout.close()
			i.p.stderr.close()
			i.p.stdin.close()
			i.p.kill()
			i.p.cleanup()
			
	def kill_orphans(self):
		if self.layer>0:
			return
		for proc in psutil.process_iter():
			try:
				self.kill_orphan(proc)
			except psutil.NoSuchProcess:
				pass

					
	def kill_orphan(self, proc):
		if proc.name() == 'python.exe' and False:#to shut down all python processes for debug, allmost never usefull.
			os.kill(proc.pid, signal.SIGTERM)
		if proc.name() == 'python.exe' and proc.cmdline()[-1]==(__main__.__file__):
			os.kill(proc.pid, signal.SIGTERM)
			if not self.kill_warned:
				print("Du to the paralle processing, only one instance of a "
					  "script can run at any time. If you want to run mulitple "
					  "instances, create multiple scripts.")
				self.kill_warned = True
			print(f"killed pid {proc.pid}" )

				
			
			
def makepath(fpath):
	fpath=os.path.join(fpath,'mp')
	os.makedirs(fpath, exist_ok=True)
	os.makedirs(fpath+'/slaves', exist_ok=True)	
	return fpath

	
class Slave():
	"""Creates a slave"""
	def __init__(self, run_parallel, n):
		"""Starts local worker"""
		self.n_nodes = n
		if run_parallel:
			path = os.path.join(os.path.dirname(__file__), "parallel_node.py")
			command = f'"{sys.executable}" -u "{path}" "{__main__.__file__}"'	
			self.p = Popen(command)
			sys.stderr = self.p.stderr
			self.t = Transact(self.p.stdout,self.p.stdin)
			
		else:
			qin, qout = Queue(), Queue()
			self.p = ThreadPopen(parallel_slave.run, (TransactThread(qout, qin),False))
			self.p.start()
			self.t = TransactThread(qin, qout)

		
	def confirm(self,slave_id, path):
		self.t.set(f'master {self.n_nodes}', self.n_nodes, path, slave_id)
		self.slave_id=slave_id
		self.p_id = self.receive()
		self.send('init_transact', (slave_id, path, self.n_nodes))
		pass

	def send(self,msg, obj):
		"""Sends msg and obj to the slave"""
		if not self.p.poll() is None:
			raise RuntimeError(f'Subprocess {self.slave_id} has unexpectedly closed')
		self.t.send((msg, obj))     

	def receive(self, q = None):
		if not self.p.poll() is None:
			raise RuntimeError(f'Subprocess {self.slave_id} has unexpectedly closed')
		answ=self.t.receive()
		if q is None:
			return answ
		q.put((answ,self.slave_id))

	
	def kill(self):
		self.p.kill()

class Transact():
	"""Local worker class"""
	def __init__(self,read, write):
		self.r = read
		self.w = write
		self.name = None
		self.slave_id = None
		self.time = time.time()
		self.f = None

	def set(self,name, n_nodes, path, slave_id):
		self.name = name
		self.n_nodes = n_nodes
		self.path = path
		self.slave_id = slave_id
		path = os.path.join(path,'debug')
		if not os.path.exists(path):
			os.mkdir(path)
		path = os.path.join(path, f'{self.name}_{self.slave_id}_{n_nodes}.txt')
		self.f = open(path, 'w',1)		

	def send(self,msg):
		self.debug_output('send',msg)
		w=getattr(self.w,'buffer',self.w)
		pickle.dump(msg,w)
		w.flush()

	def receive(self):
		r=getattr(self.r,'buffer',self.r)
		u= pickle.Unpickler(r)
		try:
			msg = u.load()
			self.debug_output('recieve', msg)
			return msg
		except pickle.UnpicklingError as e:
			r.seek(0)
			u= pickle.Unpickler(r)
			return u.load()
		except EOFError as e:
			if e.args[0]=='Ran out of input':
				raise RuntimeError(
					f"An error in {self.name} occured in sub-process {self.slave_id}." 
					f"Check the output in 'slave_errors.txt' in your working directory or "
					f"run without multiprocessing\n {datetime.datetime.now()}"
				)
			else:
				raise RuntimeError('EOFError:'+e.args[0])
			
	def debug_output(self, direction,msg):
		return
		if self.f is None:
			return
		self.f.write(f'{direction}: {self.name}, {self.slave_id}\n{time.time()-self.time}:\n{str(msg)[:30]}\ntime:{time.time()}\n')	
		self.time = time.time()		
		
class Popen(subprocess.Popen):
	def __init__(self, command):
		super().__init__(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
		self.is_pipe = True
	
	
	
class ThreadPopen(Thread):
	"""Starts local worker"""
	def __init__(self, target, args):
		super().__init__(target=target, args = args)
		self.is_pipe = False
	def kill(self):
		raise RuntimeError("No kill procedure written yet")
	def poll(self):
		if self.is_alive():
			return None
		else:
			return 1
			
class TransactThread(Transact):
	"""Thread version of Transact"""
	def __init__(self,read, write):
		self.r = read
		self.w = write
		self.name = None
		self.slave_id = None
		self.time = time.time()
		self.f = None		
		
	def set(self,name, n_nodes, path, slave_id):
		self.name = name
		self.n_nodes = n_nodes
		self.path = path
		self.slave_id = slave_id
		path = os.path.join(path,'debug')
		if not os.path.exists(path):
			os.mkdir(path)
		path = os.path.join(path, f'{self.name}_{self.slave_id}_{n_nodes}.txt')
		self.f = open(path, 'w',1)	

	def send(self,msg):
		self.debug_output('send',msg)
		self.w.put(msg) 

	def receive(self):
		msg = self.r.get()
		self.debug_output('recieve',msg)
		return msg
	
	def debug_output(self, direction,msg):
		return
		if self.f is None:
			return
		self.f.write(f'{direction}: {self.name}, {self.slave_id}\n{time.time()-self.time}:\n{str(msg)[:30]}\ntime:{time.time()}\n')	
		self.time = time.time()			
			

def write(f,txt):
	f = open(f, 'w')
	f.write(txt)
	f.flush()
	f.close()
	
def read(file):
	f = open(file, 'r')
	s = f.read()
	f.close()
	return s

def create_temp_files(cpu_count, fpath):
	number_of_files = 1

	try:#deleting old file
		f = get_file(fpath)
		folds = eval(read(f))
		for i in folds:
			os.remove(i)
	except Exception as e:
		print(e)
		a=0
	
	ftmp = []
	for i in range(number_of_files):
		f_name = gen_file_name()
		f_ = open(f_name,'w')
		f_.close()
		ftmp.append(f_name)
	write(f, str(ftmp))
	
	if number_of_files==1:
		ftmp = ftmp[0]
	return ftmp

def get_file(fpath):
	tdir = tempfile.gettempdir()
	f = os.path.join(tdir,gen_file_name(fpath))
	return f
	
def gen_file_name(seed=False):
	if not seed==False:
		f = os.path.join(os.sep.join(__file__.split(os.sep)[:-2]),os.path.join('.git','config'))
		n = len(seed)
		seed = int(hashlib.sha1(seed.encode('utf-8')).hexdigest(), 16)
		seed = os.path.getmtime(f) + seed
		random.seed(seed)
	else:
		random.seed()
	tdir = tempfile.gettempdir()
	fname = ''.join(random.choice(string.ascii_uppercase+string.ascii_lowercase + string.digits) for _ in range(20))
	fname = os.path.join(tdir,fname)
	return fname
		
		

		